from __future__ import annotations

import asyncio

import pytest

from llmling_agent import Agent


@pytest.mark.asyncio
async def test_agent_piping():
    # Use explicit names for all agents
    agent1 = Agent[None].from_callback(lambda x: f"model: {x}", name="agent1")
    transform1 = Agent[None].from_callback(
        lambda x: f"transform1: {x}", name="transform1"
    )
    transform2 = Agent[None].from_callback(
        lambda x: f"transform2: {x}", name="transform2"
    )

    pipeline = agent1 | transform1 | transform2
    result = await pipeline.run("test")

    # Check message flow
    assert result[0].message
    assert result[1].message
    assert result[2].message
    assert result[0].message.data == "model: test"
    assert result[1].message.data == "transform1: model: test"
    assert result[2].message.data == "transform2: transform1: model: test"
    assert pipeline.stats.stats.message_count == 3  # noqa: PLR2004


@pytest.mark.asyncio
async def test_agent_piping_with_monitoring():
    def callback(text: str) -> str:
        return f"model: {text}"

    agent1 = Agent[None].from_callback(callback, name="agent1")

    async def transform(text: str) -> str:
        await asyncio.sleep(0.1)  # Add a delay to ensure monitoring catches it
        return f"transform: {text}"

    pipeline = agent1 | transform

    # Get stats object directly from run_in_background
    stats = pipeline.run_in_background("test")

    # Monitor progress
    updates = []
    while pipeline.is_running:
        updates.append(len(stats))  # Track number of active connections
        await asyncio.sleep(0.01)

    _result = await pipeline.wait()
    assert updates  # Last update should show no active agents


@pytest.mark.asyncio
async def test_agent_piping_errors():
    agent1 = Agent[None].from_callback(lambda x: f"model: {x}", name="agent1")
    failing = Agent[None].from_callback(
        lambda x: exec('raise ValueError("Transform error")'),  # type: ignore
        name="failing_transform",
    )

    pipeline = agent1 | failing
    result = await pipeline.run("test")

    assert result[0].success
    assert result[0].message
    assert result[0].message.data == "model: test"
    assert not result[1].success
    assert result[1].error is not None
    assert "Transform error" in result[1].error

    assert pipeline.stats.errors


@pytest.mark.asyncio
async def test_agent_piping_async():
    async def model_callback(text: str) -> str:
        return f"model: {text}"

    agent1 = Agent[None].from_callback(model_callback, name="agent1")

    async def transform(text: str) -> str:
        return f"transform: {text}"

    pipeline = agent1 | transform
    result = await pipeline.run("test")

    assert result.successful
    assert len(result) == 2  # noqa: PLR2004
    assert result[0].message
    assert result[1].message
    assert result[0].message.data == "model: test"
    assert result[1].message.data == "transform: model: test"



"""Agent provider implementations."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Literal, Protocol, runtime_checkable

from psygnal import Signal
import tokonomics
from toprompt import to_prompt

from llmling_agent.log import get_logger
from llmling_agent.models.agents import ToolCallInfo


if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Sequence
    from contextlib import AbstractAsyncContextManager

    from tokonomics.pydanticai_cost import Usage

    from llmling_agent.agent.conversation import ConversationManager
    from llmling_agent.common_types import ModelProtocol, ModelType
    from llmling_agent.models.content import Content
    from llmling_agent.models.context import AgentContext
    from llmling_agent.models.messages import TokenCost
    from llmling_agent.tools.manager import ToolManager


logger = get_logger(__name__)


@dataclass
class ProviderResponse:
    """Raw response data from provider."""

    content: Any
    tool_calls: list[ToolCallInfo] = field(default_factory=list)
    model_name: str = ""
    cost_and_usage: TokenCost | None = None


@runtime_checkable
class StreamingResponseProtocol[TResult](Protocol):
    """Protocol for streaming responses.

    This matches PydanticAI's StreamedRunResult interface to make transition easier,
    but lives in our core package to remove the dependency.
    """

    model_name: str | None
    is_complete: bool
    formatted_content: TResult = None  # type: ignore

    def stream(self) -> AsyncIterator[TResult]:
        """Stream individual chunks as they arrive."""
        ...

    def usage(self) -> Usage:
        """Get token usage statistics if available."""
        ...


class AgentProvider[TDeps]:
    """Base class for agent providers."""

    tool_used = Signal(ToolCallInfo)
    chunk_streamed = Signal(str, str)
    model_changed = Signal(object)  # Model | None

    def __init__(
        self,
        *,
        model: str | ModelProtocol | None = None,
        name: str = "agent",
        debug: bool = False,
    ):
        super().__init__()
        self._name = name
        self._model = model
        self._tool_manager: ToolManager | None = None
        self._context: AgentContext[TDeps] | None = None
        self._conversation: ConversationManager | None = None
        self._debug = debug

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(model={self.model_name})"

    def set_model(self, model: ModelType):
        """Default no-op implementation for setting model."""

    @property
    def tool_manager(self) -> ToolManager:
        """Get tool manager."""
        if self._tool_manager is None:
            msg = "Tool manager not set"
            raise RuntimeError(msg)
        return self._tool_manager

    @tool_manager.setter
    def tool_manager(self, value: ToolManager):
        """Set tool manager."""
        self._tool_manager = value

    @property
    def context(self) -> AgentContext[TDeps]:
        """Get context."""
        if self._context is None:
            msg = "Context not set"
            raise RuntimeError(msg)
        return self._context

    @context.setter
    def context(self, value: AgentContext[TDeps]):
        """Set context."""
        self._context = value

    @property
    def conversation(self) -> ConversationManager:
        """Get conversation manager."""
        if self._conversation is None:
            msg = "Conversation manager not set"
            raise RuntimeError(msg)
        return self._conversation

    @conversation.setter
    def conversation(self, value: ConversationManager):
        """Set conversation manager."""
        self._conversation = value

    @property
    def model_name(self) -> str | None:
        """Get model name."""
        return None

    @property
    def name(self) -> str:
        """Get agent name."""
        return self._name or "agent"

    @name.setter
    def name(self, value: str | None):
        """Set agent name."""
        self._name = value or "agent"

    async def generate_response(
        self,
        *prompts: str | Content,
        message_id: str,
        result_type: type[Any] | None = None,
        model: ModelType = None,
        store_history: bool = True,
        **kwargs: Any,
    ) -> ProviderResponse:
        """Generate a response. Must be implemented by providers."""
        raise NotImplementedError

    def stream_response(
        self,
        *prompts: str | Content,
        message_id: str,
        result_type: type[Any] | None = None,
        model: ModelType = None,
        store_history: bool = True,
        **kwargs: Any,
    ) -> AbstractAsyncContextManager[StreamingResponseProtocol]:
        """Stream a response. Must be implemented by providers."""
        raise NotImplementedError

    @staticmethod
    async def format_prompts(prompts: Sequence[str | Content]) -> str:
        """Format prompts for human readability using to_prompt."""
        parts = [await to_prompt(p) for p in prompts]
        return "\n\n".join(parts)

    async def supports_feature(self, capability: Literal["vision"]) -> bool:
        """Check if provider supports a specific capability."""
        match capability:
            case "vision":
                if not self.model_name:
                    return False
                caps = await tokonomics.get_model_capabilities(self.model_name)
                return (caps and caps.supports_vision) or False
            case _:
                return False

    async def get_token_limits(self) -> tokonomics.TokenLimits | None:
        """Get token limits for the current model."""
        if not self.model_name:
            return None

        try:
            return await tokonomics.get_model_limits(self.model_name)
        except ValueError:
            logger.debug("Could not get token limits for model: %s", self.model_name)
            return None



"""Manages message flow between agents/groups."""

from __future__ import annotations

from contextlib import asynccontextmanager
from dataclasses import replace
import inspect
from typing import TYPE_CHECKING, Any, Literal, Self

from psygnal import Signal
from typing_extensions import TypeVar

from llmling_agent.log import get_logger
from llmling_agent.models.messages import ChatMessage
from llmling_agent.talk.stats import TalkStats, TeamTalkStats


if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable, Sequence
    from datetime import timedelta
    import os

    import PIL.Image
    from toprompt import AnyPromptType

    from llmling_agent.agent import AnyAgent
    from llmling_agent.common_types import AnyFilterFn, AnyTransformFn
    from llmling_agent.models.forward_targets import ConnectionType

TContent = TypeVar("TContent")
QueueStrategy = Literal["concat", "latest", "buffer"]
logger = get_logger(__name__)


class Talk[TTransmittedData]:
    """Manages message flow between agents/groups."""

    message_received = Signal(ChatMessage[TTransmittedData])  # Original message
    message_forwarded = Signal(ChatMessage[Any])  # After any transformation

    def __init__(
        self,
        source: AnyAgent[Any, TTransmittedData],
        targets: list[AnyAgent[Any, Any]],
        group: TeamTalk | None = None,
        *,
        connection_type: ConnectionType = "run",
        wait_for_connections: bool = False,
        priority: int = 0,
        delay: timedelta | None = None,
        queued: bool = False,
        queue_strategy: QueueStrategy = "latest",
        transform: AnyTransformFn | None = None,
        filter_condition: AnyFilterFn | None = None,
        stop_condition: AnyFilterFn | None = None,
        exit_condition: AnyFilterFn | None = None,
    ):
        """Initialize talk connection.

        Args:
            source: Agent sending messages
            targets: Agents receiving messages
            group: Optional group this talk belongs to
            connection_type: How to handle messages:
                - "run": Execute message as a new run in target
                - "context": Add message as context to target
                - "forward": Forward message to target's outbox
            wait_for_connections: Whether to wait for all targets to complete
            priority: Task priority (lower = higher priority)
            delay: Optional delay before processing
            queued: Whether messages should be queued for manual processing
            queue_strategy: How to process queued messages:
                - "concat": Combine all messages with newlines
                - "latest": Use only the most recent message
                - "buffer": Process all messages individually
            transform: Optional function to transform messages
            filter_condition: Optional condition for filtering messages
            stop_condition: Optional condition for disconnecting
            exit_condition: Optional condition for stopping the event loop
        """
        self.source = source
        self.targets = targets
        self.group = group
        self.priority = priority
        self.delay = delay
        self.active = True
        self.connection_type = connection_type
        self.wait_for_connections = wait_for_connections
        self.queued = queued
        self.queue_strategy = queue_strategy
        self._pending_messages: dict[str, list[ChatMessage[TTransmittedData]]] = {
            target.name: [] for target in targets
        }
        names = {t.name for t in targets}
        self._stats = TalkStats(source_name=source.name, target_names=names)
        self._transform = transform
        self._filter_condition = filter_condition
        self._stop_condition = stop_condition
        self._exit_condition = exit_condition

    def __repr__(self):
        targets = [t.name for t in self.targets]
        return f"<Talk({self.connection_type}) {self.source.name} -> {targets}>"

    async def _evaluate_condition(
        self,
        condition: Callable[..., bool | Awaitable[bool]] | None,
        message: ChatMessage[Any],
        target: AnyAgent[Any, Any],
        *,
        default_return: bool = False,
    ) -> bool:
        """Evaluate a condition with flexible parameter handling."""
        if not condition:
            return default_return

        # Get number of parameters
        sig = inspect.signature(condition)
        param_count = len(sig.parameters)

        # Call with appropriate number of arguments
        match param_count:
            case 1:
                result = condition(message)
            case 2:
                result = condition(message, target)
            case 3:
                result = condition(message, target, self.stats)
            case _:
                msg = f"Condition must take 1-3 parameters, got {param_count}"
                raise ValueError(msg)

        if inspect.isawaitable(result):
            return await result
        return result

    async def _should_route_to(
        self,
        message: ChatMessage[Any],
        target: AnyAgent[Any, Any],
    ) -> bool:
        """Determine if message should be routed to target."""
        return await self._evaluate_condition(
            self._filter_condition,
            message,
            target,
            default_return=True,
        )

    async def _handle_message(
        self,
        message: ChatMessage[TTransmittedData],
        prompt: str | None = None,
    ) -> list[ChatMessage[Any]]:
        """Handle message forwarding based on connection configuration."""
        # 1. Initial message handling
        self.source.outbox.emit(message, None)

        # 2. Early exit checks
        if not (self.active and (not self.group or self.group.active)):
            return []

        # 3. Check exit condition for any target
        for target in self.targets:
            # Exit if condition returns True
            if await self._evaluate_condition(self._exit_condition, message, target):
                raise SystemExit

        # 4. Check stop condition for any target
        for target in self.targets:
            # Stop if condition returns True
            if await self._evaluate_condition(self._stop_condition, message, target):
                self.disconnect()
                return []

        # 5. Transform if configured
        processed_message = message
        if self._transform:
            transformed = self._transform(message)
            if inspect.isawaitable(transformed):
                processed_message = await transformed
            else:
                processed_message = transformed

        # 8. Process for each target
        responses: list[ChatMessage[Any]] = []
        is_forwarded = False
        for target in self.targets:
            if await self._should_route_to(processed_message, target):
                is_forwarded = True
                if self.queued:
                    # Queue per agent
                    self._pending_messages[target.name].append(processed_message)
                    continue
                if response := await self._process_for_target(
                    processed_message, target, prompt
                ):
                    responses.append(response)
        if is_forwarded:
            self._stats = replace(
                self._stats,
                messages=[*self._stats.messages, processed_message],
            )
            # 9. Emit forwarded signal after processing
            self.message_forwarded.emit(processed_message)
        return responses

    async def _process_for_target(
        self,
        message: ChatMessage[Any],
        target: AnyAgent[Any, Any],
        prompt: AnyPromptType | PIL.Image.Image | os.PathLike[str] | None = None,
    ) -> ChatMessage[Any] | None:
        """Process message for a single target."""
        match self.connection_type:
            case "run":
                prompts: list[AnyPromptType | PIL.Image.Image | os.PathLike[str]] = [
                    message.content
                ]
                if prompt:
                    prompts.append(prompt)
                response = await target.run(*prompts)
                response.forwarded_from.append(target.name)
                target.outbox.emit(response, None)
                return response

            case "context":
                meta = {
                    "type": "forwarded_message",
                    "role": message.role,
                    "model": message.model,
                    "cost_info": message.cost_info,
                    "timestamp": message.timestamp.isoformat(),
                    "prompt": prompt,
                }

                async def add_context():
                    target.conversation.add_context_message(
                        str(message.content),
                        source=self.source.name,
                        metadata=meta,
                    )

                if self.delay is not None or self.priority != 0:
                    target.run_background(
                        add_context(),
                        priority=self.priority,
                        delay=self.delay,
                    )
                else:
                    await add_context()
                return None

            case "forward":
                if self.delay is not None or self.priority != 0:

                    async def delayed_emit():
                        target.outbox.emit(message, prompt)

                    target.run_background(
                        delayed_emit(),
                        priority=self.priority,
                        delay=self.delay,
                    )
                else:
                    target.outbox.emit(message, prompt)
                return None

    async def trigger(
        self, prompt: AnyPromptType | PIL.Image.Image | os.PathLike[str] | None = None
    ) -> list[ChatMessage[TTransmittedData]]:
        """Process queued messages."""
        if not self._pending_messages:
            return []

        match self.queue_strategy:
            case "buffer":
                results: list[ChatMessage[TTransmittedData]] = []
                # Process each agent's queue
                for target in self.targets:
                    queue = self._pending_messages[target.name]
                    for message in queue:
                        if response := await self._process_for_target(
                            message, target, prompt
                        ):
                            results.append(response)  # noqa: PERF401
                    queue.clear()
                return results

            case "latest":
                results = []
                # Get latest message for each agent
                for target in self.targets:
                    queue = self._pending_messages[target.name]
                    if queue:
                        latest = queue[-1]
                        if response := await self._process_for_target(
                            latest, target, prompt
                        ):
                            results.append(response)
                        queue.clear()
                return results

            case "concat":
                results = []
                # Concat messages per agent
                for target in self.targets:
                    queue = self._pending_messages[target.name]
                    if not queue:
                        continue

                    base = queue[-1]
                    contents = [str(m.content) for m in queue]
                    meta = {
                        **base.metadata,
                        "merged_count": len(queue),
                        "queue_strategy": self.queue_strategy,
                    }
                    content = "\n\n".join(contents)
                    merged = replace(base, content=content, metadata=meta)  # type: ignore

                    if response := await self._process_for_target(merged, target, prompt):
                        results.append(response)
                    queue.clear()

                return results
            case _:
                msg = f"Invalid queue strategy: {self.queue_strategy}"
                raise ValueError(msg)

    def when(self, condition: AnyFilterFn) -> Self:
        """Add condition for message forwarding."""
        self._filter_condition = condition
        return self

    @asynccontextmanager
    async def paused(self):
        """Temporarily set inactive."""
        previous = self.active
        self.active = False
        try:
            yield self
        finally:
            self.active = previous

    def disconnect(self):
        """Permanently disconnect the connection."""
        self.active = False

    @property
    def stats(self) -> TalkStats:
        """Get current connection statistics."""
        return self._stats


class TeamTalk(list["Talk | TeamTalk"]):
    """Group of connections with aggregate operations."""

    def __init__(self, talks: Sequence[Talk | TeamTalk]):
        super().__init__(talks)
        self._filter_condition: AnyFilterFn | None = None
        self.active = True

    def __repr__(self):
        return f"TeamTalk({list(self)})"

    @property
    def targets(self) -> list[AnyAgent[Any, Any]]:
        """Get all targets from all connections."""
        return [t for talk in self for t in talk.targets]

    async def _handle_message(self, message: ChatMessage[Any], prompt: str | None = None):
        for talk in self:
            await talk._handle_message(message, prompt)

    async def trigger(
        self, prompt: AnyPromptType | PIL.Image.Image | os.PathLike[str] | None = None
    ) -> list[ChatMessage]:
        messages = []
        for talk in self:
            messages.extend(await talk.trigger(prompt))
        return messages

    @classmethod
    def from_agents(
        cls,
        agents: Sequence[AnyAgent[Any, Any]],
        targets: list[AnyAgent[Any, Any]] | None = None,
    ) -> Self:
        """Create TeamTalk from a collection of agents."""
        return cls([Talk(agent, targets or []) for agent in agents])

    @asynccontextmanager
    async def paused(self):
        """Temporarily set inactive."""
        previous = self.active
        self.active = False
        try:
            yield self
        finally:
            self.active = previous

    def has_active_talks(self) -> bool:
        """Check if any contained talks are active."""
        return any(talk.active for talk in self)

    def get_active_talks(self) -> list[Talk | TeamTalk]:
        """Get list of currently active talks."""
        return [talk for talk in self if talk.active]

    @property
    def stats(self) -> TeamTalkStats:
        """Get aggregated statistics for all connections."""
        return TeamTalkStats(stats=[talk.stats for talk in self])

    def when(self, condition: AnyFilterFn) -> Self:
        """Add condition to all connections in group."""
        for talk in self:
            talk.when(condition)
        return self

    def disconnect(self):
        """Disconnect all connections in group."""
        for talk in self:
            talk.disconnect()



from __future__ import annotations

from collections.abc import Callable, Sequence
from contextlib import AsyncExitStack, asynccontextmanager
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, overload

from psygnal.containers import EventedList
from typing_extensions import TypeVar

from llmling_agent.agent.connection import ConnectionManager
from llmling_agent.delegation.teamrun import ExecutionMode, TeamRun
from llmling_agent.log import get_logger
from llmling_agent.talk import QueueStrategy, TeamTalk
from llmling_agent.utils.inspection import has_return_type
from llmling_agent.utils.tasks import TaskManagerMixin
from llmling_agent_providers.base import StreamingResponseProtocol


logger = get_logger(__name__)

if TYPE_CHECKING:
    from collections.abc import AsyncIterator
    from datetime import timedelta
    import os

    import PIL.Image
    from tokonomics.pydanticai_cost import Usage
    from toprompt import AnyPromptType

    from llmling_agent.agent import AnyAgent
    from llmling_agent.common_types import AgentName, AnyTransformFn, AsyncFilterFn
    from llmling_agent.models.forward_targets import ConnectionType
    from llmling_agent.models.messages import ChatMessage, TeamResponse
    from llmling_agent.models.providers import ProcessorCallback


TDeps = TypeVar("TDeps", default=None)


class Team[TDeps](TaskManagerMixin):
    """Group of agents that can execute together."""

    def __init__(
        self,
        agents: Sequence[AnyAgent[TDeps, Any]],
        *,
        shared_prompt: str | None = None,
        name: str | None = None,
    ):
        self.agents = EventedList(list(agents))
        self.shared_prompt = shared_prompt
        self.name = name or " & ".join([i.name for i in agents])
        self.connections = ConnectionManager(self)
        self.team_talk = TeamTalk.from_agents(self.agents)

    def __repr__(self) -> str:
        """Create a readable representation of the team."""
        members = ", ".join(agent.name for agent in self.agents)
        name = f" ({self.name})" if self.name else ""
        return f"Team[{len(self.agents)}]{name}: {members}"

    def __contains__(self, item: Any) -> bool:
        """Check if an agent is part of the team."""
        return item in self.agents

    def __iter__(self):
        """Iterate over team members."""
        return iter(self.agents)

    def __len__(self) -> int:
        """Get number of team members."""
        return len(self.agents)

    def __getitem__(self, index_or_name: int | AgentName) -> AnyAgent[TDeps, Any]:
        """Get team member by index or name."""
        if isinstance(index_or_name, str):
            return next(agent for agent in self.agents if agent.name == index_or_name)
        return self.agents[index_or_name]

    @overload
    def __and__(self, other: Team[None]) -> Team[None]: ...

    @overload
    def __and__(self, other: Team[TDeps]) -> Team[TDeps]: ...

    @overload
    def __and__(self, other: Team[Any]) -> Team[Any]: ...

    @overload
    def __and__(self, other: AnyAgent[TDeps, Any]) -> Team[TDeps]: ...

    @overload
    def __and__(self, other: AnyAgent[Any, Any]) -> Team[Any]: ...

    def __and__(self, other: Team[Any] | AnyAgent[Any, Any]) -> Team[Any]:
        """Combine teams, preserving type safety for same types."""
        # Combine agents

        match other:
            case Team():
                combined_agents = [*self.agents, *other.agents]
                combined_prompts = []
                if self.shared_prompt:
                    combined_prompts.append(self.shared_prompt)
                if other.shared_prompt:
                    combined_prompts.append(other.shared_prompt)
                shared = "\n".join(combined_prompts) if combined_prompts else None
                return Team(agents=combined_agents, shared_prompt=shared)
            case _:  # AnyAgent case
                # Keep same deps if types match
                agents = [*self.agents, other]
                return Team(agents=agents, shared_prompt=self.shared_prompt)

    def __or__(
        self, other: AnyAgent[Any, Any] | ProcessorCallback[Any] | Team[Any]
    ) -> TeamRun[TDeps]:
        """Create a pipeline using | operator.

        Example:
            pipeline = team | transform | other_team  # Sequential processing
        """
        from llmling_agent.agent import Agent, StructuredAgent

        match other:
            case Team():
                # Create sequential execution with all agents
                team = Team([*self.agents, *other.agents])
                execution = TeamRun(team, mode="sequential")
            case Callable():
                # Convert callable to agent and add to pipeline
                from llmling_agent import Agent
                from llmling_agent_providers.callback import CallbackProvider

                provider = CallbackProvider(other)
                new_agent = Agent(provider=provider, name=other.__name__)
                execution = TeamRun(Team([*self.agents, new_agent]), mode="sequential")
            case Agent() | StructuredAgent():  # Agent case
                execution = TeamRun(Team([*self.agents, other]), mode="sequential")
            case _:
                msg = f"Invalid pipeline element: {other}"
                raise ValueError(msg)

        # Setup connections for sequential processing
        for i in range(len(execution.team.agents) - 1):
            current = execution.team.agents[i]
            next_agent = execution.team.agents[i + 1]
            current.pass_results_to(next_agent)

        return execution

    def __rshift__(
        self, other: AnyAgent[Any, Any] | Team[Any] | ProcessorCallback[Any]
    ) -> TeamTalk:
        """Connect group to target agent(s).

        Returns:
            - list[Talk] when connecting to single agent (one Talk per source)
            - list[TeamTalk] when connecting to team (one TeamTalk per source)
        """
        return self.pass_results_to(other)

    def pass_results_to(
        self,
        other: AnyAgent[Any, Any] | Team[Any] | ProcessorCallback[Any],
        *,
        connection_type: ConnectionType = "run",
        priority: int = 0,
        delay: timedelta | None = None,
        queued: bool = False,
        queue_strategy: QueueStrategy = "latest",
        transform: AnyTransformFn | None = None,
        filter_condition: AsyncFilterFn | None = None,
        stop_condition: AsyncFilterFn | None = None,
        exit_condition: AsyncFilterFn | None = None,
    ) -> TeamTalk:
        """Forward results to another agent or all agents in a team."""
        from llmling_agent.agent import Agent, StructuredAgent

        if callable(other):
            if has_return_type(other, str):
                other = Agent.from_callback(other)
            else:
                other = StructuredAgent.from_callback(other)
        return self.connections.connect_group_to(
            other,
            connection_type=connection_type,
            priority=priority,
            delay=delay,
            queued=queued,
            queue_strategy=queue_strategy,
            transform=transform,
            filter_condition=filter_condition,
            stop_condition=stop_condition,
            exit_condition=exit_condition,
        )

    def monitored(self, mode: ExecutionMode) -> TeamRun[TDeps]:
        """Create a monitored execution."""
        return TeamRun(self, mode)

    async def run_parallel(
        self,
        *prompts: AnyPromptType | PIL.Image.Image | os.PathLike[str] | None,
    ) -> TeamResponse:
        """Run all agents in parallel."""
        execution = TeamRun(self, "parallel")
        return await execution.run(*prompts)

    async def run_sequential(
        self,
        *prompts: AnyPromptType | PIL.Image.Image | os.PathLike[str] | None,
    ) -> TeamResponse:
        """Run agents one after another."""
        execution = TeamRun(self, "sequential")
        return await execution.run(*prompts)

    async def chain(
        self,
        *prompts: AnyPromptType | PIL.Image.Image | os.PathLike[str] | None,
        require_all: bool = True,
    ) -> ChatMessage:
        """Pass message through the chain of team members.

        Each agent processes the result of the previous one.

        Args:
            prompts: Initial messages to process
            require_all: If True, all agents must succeed

        Returns:
            Final processed message

        Raises:
            ValueError: If chain breaks and require_all=True
        """
        current_message = prompts

        for agent in self.agents:
            try:
                result = await agent.run(*current_message)
                current_message = (result.content,)
            except Exception as e:
                if require_all:
                    msg = f"Chain broken at {agent.name}: {e}"
                    raise ValueError(msg) from e
                logger.warning("Chain handler %s failed: %s", agent.name, e)

        return result

    @asynccontextmanager
    async def chain_stream(
        self,
        *prompts: AnyPromptType | PIL.Image.Image | os.PathLike[str] | None,
        require_all: bool = True,
    ) -> AsyncIterator[StreamingResponseProtocol]:
        """Stream results through chain of team members."""
        async with AsyncExitStack() as stack:
            streams: list[StreamingResponseProtocol[str]] = []
            current_message = prompts

            # Set up all streams
            for agent in self.agents:
                try:
                    stream = await stack.enter_async_context(
                        agent.run_stream(*current_message)
                    )
                    streams.append(stream)
                    # Wait for complete response for next agent
                    async for chunk in stream.stream():
                        current_message = chunk
                        if stream.is_complete:
                            current_message = (stream.formatted_content,)  # type: ignore
                            break
                except Exception as e:
                    if require_all:
                        msg = f"Chain broken at {agent.name}: {e}"
                        raise ValueError(msg) from e
                    logger.warning("Chain handler %s failed: %s", agent.name, e)

            # Create a stream-like interface for the chain
            class ChainStream(StreamingResponseProtocol[str]):
                def __init__(self):
                    self.streams = streams
                    self.current_stream_idx = 0
                    self.is_complete = False
                    self.model_name = None

                def usage(self) -> Usage:
                    @dataclass
                    class Usage:
                        total_tokens: int | None
                        request_tokens: int | None
                        response_tokens: int | None

                    return Usage(0, 0, 0)

                async def stream(self) -> AsyncIterator[str]:  # type: ignore
                    for idx, stream in enumerate(self.streams):
                        self.current_stream_idx = idx
                        async for chunk in stream.stream():
                            yield chunk
                            if idx == len(self.streams) - 1 and stream.is_complete:
                                self.is_complete = True

            yield ChainStream()

    async def distribute(
        self,
        content: str,
        *,
        tools: list[str] | None = None,
        resources: list[str] | None = None,
    ) -> None:
        """Distribute content and capabilities to all team members.

        This method provides content and optional capabilities to every agent in the team.

        Args:
            content: Text content to add as context for all agents
            tools: Optional list of tools to register with each agent.
                  Can be import paths or callables.
            resources: Optional list of resources to load into each agent's context.
                      Can be paths, URLs, or resource configurations.

        """
        for agent in self.agents:
            # Add context message
            agent.conversation.add_context_message(content, source="distribution")

            # Register tools if provided
            if tools:
                for tool in tools:
                    agent.tools.register_tool(tool)

            # Load resources if provided
            if resources:
                for resource in resources:
                    await agent.conversation.load_context_source(resource)



"""Agent chat session management."""

from llmling_agent.chat_session.base import AgentPoolView
from llmling_agent.chat_session.exceptions import ChatSessionError

__all__ = ["AgentPoolView", "ChatSessionError"]



"""Rename chat screen. Credits to Elia (https://github.com/darrenburns/elia)."""

from typing import ClassVar

from textual import on
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Vertical
from textual.screen import ModalScreen
from textual.widgets import Input


class RenameChat(ModalScreen[str]):
    BINDINGS: ClassVar = [
        Binding("escape", "app.pop_screen", "Cancel", key_display="esc"),
        Binding("enter", "app.pop_screen", "Save"),
    ]

    def compose(self) -> ComposeResult:
        with Vertical():
            title_input = Input(placeholder="Enter a title...")
            title_input.border_subtitle = "[[white]enter[/]] Save  [[white]esc[/]] Cancel"
            yield title_input

    @on(Input.Submitted)
    def close_screen(self, event: Input.Submitted):
        self.dismiss(event.value)



"""Event sources for LLMling agent."""

from __future__ import annotations

from abc import abstractmethod
from dataclasses import dataclass, field
from datetime import datetime
from typing import Annotated, Any, Literal, Self

from pydantic import BaseModel, ConfigDict, Field, SecretStr


DEFAULT_TEMPLATE = """
{%- if include_timestamp %}at {{ timestamp }}{% endif %}
Event from {{ source }}:
{%- if include_metadata %}
Metadata:
{% for key, value in metadata.items() %}
{{ key }}: {{ value }}
{% endfor %}
{% endif %}
{{ content }}
"""


@dataclass(frozen=True, kw_only=True)
class EventData:
    """Base class for event data."""

    source: str
    timestamp: datetime = field(default_factory=datetime.now)
    metadata: dict[str, Any] = field(default_factory=dict)

    @classmethod
    def create(cls, source: str, **kwargs: Any) -> Self:
        """Create event with current timestamp."""
        return cls(source=source, timestamp=datetime.now(), **kwargs)

    @abstractmethod
    def to_prompt(self) -> str:
        """Convert event to agent prompt."""

    async def format(self, config: EventSourceConfig) -> str:
        """Wraps core message with configurable template."""
        from jinja2 import Environment

        env = Environment(trim_blocks=True, lstrip_blocks=True)
        template = env.from_string(config.template)

        return template.render(
            source=self.source,
            content=self.to_prompt(),  # Use the core message
            metadata=self.metadata,
            timestamp=self.timestamp,
            include_metadata=config.include_metadata,
            include_timestamp=config.include_timestamp,
        )


@dataclass(frozen=True, kw_only=True)
class UIEvent(EventData):
    """Event triggered through UI interaction."""

    type: Literal["command", "message", "agent_command", "agent_message"]
    """Type of UI interaction that triggered this event."""

    content: str
    """The actual content (command string, voice command, etc.)."""

    args: list[str] = field(default_factory=list)
    """Additional arguments for the interaction."""

    kwargs: dict[str, Any] = field(default_factory=dict)
    """Additional options/parameters."""

    agent_name: str | None = None
    """Target agent for @agent messages/commands."""

    def to_prompt(self) -> str:
        """Convert event to agent prompt."""
        match self.type:
            case "command":
                args_str = " ".join(self.args)
                kwargs_str = " ".join(f"--{k}={v}" for k, v in self.kwargs.items())
                return f"UI Command: /{self.content} {args_str} {kwargs_str}"
            case "shortcut" | "gesture":
                return f"UI Action: {self.content}"
            case "voice":
                return f"Voice Command: {self.content}"
            case _:
                raise ValueError(self.type)


class EventSourceConfig(BaseModel):
    """Base configuration for event sources."""

    type: str = Field(init=False)
    """Discriminator field for event source types."""

    name: str
    """Unique identifier for this event source."""

    enabled: bool = True
    """Whether this event source is active."""

    template: str = DEFAULT_TEMPLATE
    """Jinja2 template for formatting events."""

    include_metadata: bool = True
    """Control metadata visibility in template."""

    include_timestamp: bool = True
    """Control timestamp visibility in template."""

    model_config = ConfigDict(frozen=True, use_attribute_docstrings=True)


class FileWatchConfig(EventSourceConfig):
    """File watching event source.

    Monitors file system changes and triggers agent actions when:
    - Files are created
    - Files are modified
    - Files are deleted
    """

    type: Literal["file"] = Field("file", init=False)
    """Type discriminator for file watch sources."""

    paths: list[str]
    """Paths or patterns to watch for changes."""

    extensions: list[str] | None = None
    """File extensions to monitor (e.g. ['.py', '.md'])."""

    ignore_paths: list[str] | None = None
    """Paths or patterns to ignore."""

    recursive: bool = True
    """Whether to watch subdirectories."""

    debounce: int = 1600
    """Minimum time (ms) between trigger events."""


class WebhookConfig(EventSourceConfig):
    """Webhook event source.

    Listens for HTTP requests and triggers agent actions when:
    - POST requests are received at the configured endpoint
    - Request content matches any defined filters
    """

    type: Literal["webhook"] = Field("webhook", init=False)
    """Type discriminator for webhook sources."""

    port: int
    """Port to listen on."""

    path: str
    """URL path to handle requests."""

    secret: str | None = None
    """Optional secret for request validation."""


class TimeEventConfig(EventSourceConfig):
    """Time-based event source configuration."""

    type: Literal["time"] = Field("time", init=False)
    """Type discriminator for time events."""

    schedule: str
    """Cron expression for scheduling (e.g. '0 9 * * 1-5' for weekdays at 9am)"""

    prompt: str
    """Prompt to send to the agent when the schedule triggers."""

    timezone: str | None = None
    """Timezone for schedule (defaults to system timezone)"""

    skip_missed: bool = False
    """Whether to skip executions missed while agent was inactive"""


class EmailConfig(EventSourceConfig):
    """Email event source configuration.

    Monitors an email inbox for new messages and converts them to events.
    """

    type: Literal["email"] = Field("email", init=False)
    """Type discriminator for email sources."""

    host: str = Field(description="IMAP server hostname")
    """IMAP server hostname (e.g. 'imap.gmail.com')"""

    port: int = Field(default=993)
    """Server port (defaults to 993 for IMAP SSL)"""

    username: str
    """Email account username/address"""

    password: SecretStr
    """Account password or app-specific password"""

    folder: str = Field(default="INBOX")
    """Folder/mailbox to monitor"""

    ssl: bool = Field(default=True)
    """Whether to use SSL/TLS connection"""

    check_interval: int = Field(
        default=60, gt=0, description="Seconds between inbox checks"
    )
    """How often to check for new emails (in seconds)"""

    mark_seen: bool = Field(default=True)
    """Whether to mark processed emails as seen"""

    filters: dict[str, str] = Field(
        default_factory=dict, description="Email filtering criteria"
    )
    """Filtering rules for emails (subject, from, etc)"""

    max_size: int | None = Field(default=None, description="Maximum email size in bytes")
    """Size limit for processed emails"""


EventConfig = Annotated[
    FileWatchConfig | WebhookConfig | EmailConfig | TimeEventConfig,
    Field(discriminator="type"),
]



"""Utilities for database storage."""

from __future__ import annotations

from typing import TYPE_CHECKING

from pydantic_ai.messages import (
    ModelMessage,
    ModelRequest,
    ModelResponse,
    SystemPromptPart,
    TextPart,
    UserPromptPart,
)


if TYPE_CHECKING:
    from llmling_agent_storage.sql_provider.models import Message


def db_message_to_pydantic_ai_message(msg: Message) -> ModelMessage:
    """Convert a database message to a pydantic-ai message."""
    match msg.role:
        case "user":
            return ModelRequest(parts=[UserPromptPart(content=msg.content)])
        case "assistant":
            return ModelResponse(parts=[TextPart(content=msg.content)])
        case "system":
            return ModelRequest(parts=[SystemPromptPart(content=msg.content)])
    error_msg = f"Unknown message role: {msg.role}"
    raise ValueError(error_msg)


def parse_model_info(model: str | None) -> tuple[str | None, str | None]:
    """Parse model string into provider and name.

    Args:
        model: Full model string (e.g., "openai:gpt-4", "anthropic/claude-2")

    Returns:
        Tuple of (provider, name)
    """
    if not model:
        return None, None

    # Try splitting by ':' or '/'
    parts = model.split(":") if ":" in model else model.split("/")

    if len(parts) == 2:  # noqa: PLR2004
        provider, name = parts
        return provider.lower(), name

    # No provider specified, try to infer
    name = parts[0]
    if name.startswith(("gpt-", "text-", "dall-e")):
        return "openai", name
    if name.startswith("claude"):
        return "anthropic", name
    if name.startswith(("llama", "mistral")):
        return "meta", name

    return None, name



"""Event handling for LLMling agents."""

from llmling_agent.events.manager import EventManager
from llmling_agent.events.sources import EventConfig, FileWatchConfig, WebhookConfig

__all__ = ["EventConfig", "EventManager", "FileWatchConfig", "WebhookConfig"]



"""Team execution management and monitoring."""

from __future__ import annotations

import asyncio
from collections.abc import AsyncIterator, Callable
from dataclasses import dataclass, field
from datetime import datetime
from itertools import pairwise
from time import perf_counter
from typing import TYPE_CHECKING, Any, Literal

from llmling_agent.log import get_logger
from llmling_agent.models.messages import AgentResponse, ChatMessage, TeamResponse
from llmling_agent.talk.talk import Talk, TeamTalk
from llmling_agent.utils.tasks import TaskManagerMixin


if TYPE_CHECKING:
    import os

    import PIL.Image
    from toprompt import AnyPromptType

    from llmling_agent.agent import AnyAgent
    from llmling_agent.agent.agent import Agent
    from llmling_agent.delegation import Team


logger = get_logger(__name__)

ExecutionMode = Literal["parallel", "sequential"]
"""The execution mode for a TeamRun."""


@dataclass(frozen=True, kw_only=True)
class ExtendedTeamTalk(TeamTalk):
    """TeamTalk that also provides TeamRunStats interface."""

    errors: list[tuple[str, str, datetime]] = field(default_factory=list)

    def add_error(self, agent: str, error: str):
        """Track errors from AgentResponses."""
        self.errors.append((agent, error, datetime.now()))

    @property
    def error_log(self) -> list[tuple[str, str, datetime]]:
        """Errors from failed responses."""
        return self.errors


class TeamRun[TDeps](TaskManagerMixin):
    """Handles team operations with monitoring."""

    def __init__(
        self,
        team: Team[TDeps],
        mode: ExecutionMode,
    ):
        super().__init__()
        self.team = team
        self.mode = mode
        self._team_talk = ExtendedTeamTalk()
        self._main_task: asyncio.Task[TeamResponse] | None = None

    def __or__(self, other: Agent | Callable | Team | TeamRun) -> TeamRun:
        from llmling_agent import Agent, Team
        from llmling_agent_providers.callback import CallbackProvider

        match other:
            case Agent():
                self.team.agents.append(other)
            case Callable():
                provider = CallbackProvider(other)
                position = len(self.team.agents) + 1
                name = f"{other.__name__}_{position}"
                new_agent = Agent(provider=provider, name=name)
                self.team.agents.append(new_agent)
            case Team():
                # Flatten team
                self.team.agents.extend(other.agents)
            case TeamRun():
                # Merge executions
                self.team.agents.extend(other.team.agents)
        return self

    async def run(
        self,
        *prompts: AnyPromptType | PIL.Image.Image | os.PathLike[str] | None,
        **kwargs: Any,
    ) -> TeamResponse:
        """Start execution with optional monitoring."""
        self._team_talk = ExtendedTeamTalk()
        try:
            match self.mode:
                case "parallel":
                    return await self._run_parallel(*prompts)
                case "sequential":
                    return await self._run_sequential(*prompts)
                case _:
                    msg = f"Invalid mode: {self.mode}"
                    raise ValueError(msg)
        finally:
            pass

    def run_in_background(
        self,
        *prompts: AnyPromptType | PIL.Image.Image | os.PathLike[str] | None,
        **kwargs: Any,
    ) -> ExtendedTeamTalk:
        if self._main_task:
            msg = "Execution already running"
            raise RuntimeError(msg)
        coro = self.run(*prompts, **kwargs)
        self._main_task = self.create_task(coro, name="main_execution")
        return self._team_talk

    @property
    def is_running(self) -> bool:
        """Whether execution is currently running."""
        return bool(self._main_task and not self._main_task.done())

    async def wait(self) -> TeamResponse:
        if not self._main_task:
            msg = "No execution running"
            raise RuntimeError(msg)
        try:
            return await self._main_task
        finally:
            await self.cleanup_tasks()

    async def cancel(self) -> None:
        """Cancel execution and cleanup."""
        if self._main_task:
            self._main_task.cancel()
        await self.cleanup_tasks()

    @property
    def stats(self) -> ExtendedTeamTalk:
        """Get current execution statistics."""
        return self._team_talk

    async def _run_parallel(
        self,
        *prompt: AnyPromptType | PIL.Image.Image | os.PathLike[str],
    ) -> TeamResponse:
        """Execute in parallel mode."""
        start_time = datetime.now()
        final_prompt = list(prompt)
        if self.team.shared_prompt:
            final_prompt.insert(0, self.team.shared_prompt)

        async def run_agent(agent: AnyAgent[TDeps, Any]) -> AgentResponse[Any]:
            try:
                start = perf_counter()
                message = await agent.run(*final_prompt)
                timing = perf_counter() - start
                return AgentResponse(agent.name, message=message, timing=timing)
            except Exception as e:  # noqa: BLE001
                msg = ChatMessage(content="", role="assistant")
                self._team_talk.add_error(agent.name, str(e))
                return AgentResponse(agent_name=agent.name, message=msg, error=str(e))

        responses = await asyncio.gather(*[run_agent(a) for a in self.team.agents])
        return TeamResponse(responses, start_time)

    async def run_iter(
        self,
        *prompt: AnyPromptType | PIL.Image.Image | os.PathLike[str],
    ) -> AsyncIterator[Talk[Any] | AgentResponse[Any]]:
        try:
            first = self.team[0]
            connections = [
                source.pass_results_to(target, queued=True)
                for source, target in pairwise(self.team)
            ]
            for conn in connections:
                self._team_talk.append(conn)

            start = perf_counter()
            message = await first.run(*prompt)
            timing = perf_counter() - start
            response = AgentResponse[Any](
                agent_name=first.name, message=message, timing=timing
            )
            yield response
            for connection in connections:
                target = connection.targets[0]
                target_name = target.name
                yield connection  # pyright: ignore
                try:
                    start = perf_counter()
                    messages = await connection.trigger()
                    # If this is the last agent
                    if target == self.team.agents[-1]:
                        # Create Talk for stats collection only
                        last_talk = Talk[Any](target, [], connection_type="run")
                        # Add its message to the Talk's stats
                        if response.message:
                            last_talk.stats.messages.append(response.message)
                        # Add Talk to TeamTalk
                        self._team_talk.append(last_talk)
                    timing = perf_counter() - start
                    response = AgentResponse[Any](
                        agent_name=target_name, message=messages[0], timing=timing
                    )
                    yield response

                except Exception as e:  # noqa: BLE001
                    self._team_talk.add_error(connection.targets[0].name, str(e))
                    msg = ChatMessage(content="", role="assistant")
                    response = AgentResponse[Any](
                        agent_name=target_name, message=msg, error=str(e)
                    )
                    yield response
        finally:
            for connection in connections:
                connection.disconnect()

    async def _run_sequential(
        self, *prompt: AnyPromptType | PIL.Image.Image | os.PathLike[str]
    ) -> TeamResponse:
        """Execute in sequential mode."""
        start_time = datetime.now()
        final_prompt = list(prompt)
        if self.team.shared_prompt:
            final_prompt.insert(0, self.team.shared_prompt)

        responses = [
            i async for i in self.run_iter(*final_prompt) if isinstance(i, AgentResponse)
        ]
        return TeamResponse(responses, start_time)


if __name__ == "__main__":
    import asyncio

    from llmling_agent import AgentPool

    async def main():
        async with AgentPool[None]() as pool:
            # Create three agents with different roles
            agent1 = await pool.add_agent(
                "analyzer",
                system_prompt="You analyze text and find key points.",
                model="openai:gpt-4o-mini",
            )
            agent2 = await pool.add_agent(
                "summarizer",
                system_prompt="You create concise summaries.",
                model="openai:gpt-4o-mini",
            )
            agent3 = await pool.add_agent(
                "critic",
                system_prompt="You evaluate and critique summaries.",
                model="openai:gpt-4o-mini",
            )

            # Create team and get monitored execution
            team = agent1 & agent2 & agent3
            run = team.monitored(mode="sequential")

            text = "The quick brown fox jumps over the lazy dog."
            print(f"\nProcessing text: {text}\n")

            # Start run and get stats object (ExtendedTeamTalk)
            stats = run.run_in_background(text)

            # Poll stats while running
            while run.is_running:
                print("\nCurrent status:")
                print(f"Number of active connections: {len(stats)}")
                print("Errors:", len(stats.errors))
                await asyncio.sleep(0.5)

            # Wait for completion and get results
            result = await run.wait()
            print("\nFinal Results:")
            for resp in result:
                print(f"\n{resp.agent_name}:")
                print("-" * 40)
                print(resp.message)

    asyncio.run(main())



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

import logfire
from slashed.textual_adapter import CommandInput
from textual import on
from textual.app import App, ComposeResult
from textual.containers import Container
from textual.widgets import Header

from llmling_agent import Agent
from llmling_agent.models import ChatMessage
from llmling_agent.utils.tasks import TaskManagerMixin
from llmling_agent_commands import get_commands
from llmling_textual.widget import ChatView


if TYPE_CHECKING:
    from llmling_agent.agent import AnyAgent


logger = logging.getLogger(__name__)


class ChatApp(App, TaskManagerMixin):
    """Chat application with command support."""

    def __init__(self):
        super().__init__()
        self._agent: AnyAgent[Any, Any] | None = None
        self._agent_cm: Any = None

    def compose(self) -> ComposeResult:
        yield Header()
        yield Container(ChatView(), id="chat-output")
        yield CommandInput(
            output_id="chat-output",
            context_data=self._agent,
            status_id="status-area",
        )

    async def on_mount(self):
        """Initialize agent when app starts."""
        path = "src/llmling_agent/config_resources/agents.yml"
        self._agent_cm = Agent.open_agent(path, "url_opener")
        self._agent = await self._agent_cm.__aenter__()
        assert self._agent
        logger.info("Agent initialized: %s", self._agent.name)
        command_input = self.query_one(CommandInput)
        for command in get_commands():
            command_input.store.register_command(command)

    async def on_unmount(self):
        """Clean up tasks and agent."""
        await self.complete_tasks(cancel=True)

        if self._agent_cm is not None:
            await self._agent_cm.__aexit__(None, None, None)

    @logfire.instrument("Got input: {event.text}")
    @on(CommandInput.InputSubmitted)
    async def handle_submit(self, event: CommandInput.InputSubmitted):
        """Handle regular input submission."""
        self.create_task(
            self.handle_chat_message(event.text), name=f"chat_message_{event.text[:10]}"
        )

    @logfire.instrument("Processing message: {text}")
    async def handle_chat_message(self, text: str):
        """Process normal chat messages."""
        if not self._agent:
            logger.error("No agent available!")
            return
        chat = self.query_one(ChatView)

        # Add user message
        await chat.add_message(ChatMessage[str](role="user", content=text, name="User"))

        # Create empty assistant message, will get populated for streaming.
        name = self._agent.model_name
        msg = ChatMessage[str](role="assistant", content="", name="Assistant", model=name)
        widget = await chat.add_message(msg)

        try:
            # Stream the response from the agent
            logger.debug("Starting stream")
            async with self._agent.run_stream(text) as stream:
                full_content = ""
                logger.debug("Stream opened, waiting for chunks")
                async for chunk in stream.stream():
                    # Get just the new content by removing the overlap
                    new_content = str(chunk).removeprefix(full_content)
                    logger.debug("Got new content: %r", new_content)

                    full_content += new_content
                    widget.update_content(full_content)
                    widget.refresh()

                logger.debug("Stream completed with final content: %r", full_content)
        except Exception:
            logger.exception("Error streaming response")
            widget.update_content("Error processing response")


if __name__ == "__main__":
    logging.basicConfig(
        level=logging.DEBUG,
        filename="chat_app.log",
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )
    app = ChatApp()
    app.run()



"""Talk classes."""

from llmling_agent.talk.stats import TalkStats, TeamTalkStats
from llmling_agent.talk.talk import Talk, TeamTalk, QueueStrategy

__all__ = [
    "QueueStrategy",
    "Talk",
    "TalkStats",
    "TeamTalk",
    "TeamTalkStats",
]



"""Agent configuration and creation."""

from llmling_agent.models import AgentsManifest, AgentConfig, AgentContext
from llmling_agent.agent import Agent, StructuredAgent

from llmling_agent.delegation import AgentPool, Team, TeamRun
from dotenv import load_dotenv
from llmling_agent.models.messages import ChatMessage
from llmling_agent.chat_session.base import AgentPoolView

__version__ = "0.95.1"

load_dotenv()

__all__ = [
    "Agent",
    "AgentConfig",
    "AgentContext",
    "AgentPool",
    "AgentPoolView",
    "AgentsManifest",
    "ChatMessage",
    "StructuredAgent",
    "Team",
    "TeamRun",
]



"""Manages message flow between agents/groups."""

from __future__ import annotations

from contextlib import asynccontextmanager
from dataclasses import replace
from typing import TYPE_CHECKING, Any, overload

from psygnal import Signal

from llmling_agent.log import get_logger
from llmling_agent.talk import Talk, TeamTalk, TeamTalkStats


if TYPE_CHECKING:
    from datetime import timedelta

    from llmling_agent.agent import AnyAgent
    from llmling_agent.common_types import AgentName, AnyFilterFn, AnyTransformFn
    from llmling_agent.delegation.team import Team
    from llmling_agent.models.forward_targets import ConnectionType
    from llmling_agent.models.messages import ChatMessage
    from llmling_agent.talk.talk import QueueStrategy

logger = get_logger(__name__)


class ConnectionManager:
    """Manages connections for both Agents and Teams."""

    agent_connected = Signal(object)  # Agent
    connection_added = Signal(Talk)  # Agent

    def __init__(self, owner: AnyAgent[Any, Any] | Team[Any]):
        self.owner = owner
        self._connections: list[Talk | TeamTalk] = []
        self._wait_states: dict[AgentName, bool] = {}

    def __repr__(self):
        return f"ConnectionManager({self.owner})"

    def set_wait_state(self, target: AnyAgent[Any, Any] | AgentName, wait: bool = True):
        """Set waiting behavior for target."""
        target_name = target if isinstance(target, str) else target.name
        self._wait_states[target_name] = wait

    async def wait_for_connections(self, _seen: set[AgentName] | None = None):
        """Wait for this agent and all connected agents to complete their tasks."""
        seen: set[AgentName] = _seen or {self.owner.name}  # type: ignore

        # Wait for our own tasks
        await self.owner.complete_tasks()

        # Wait for connected agents
        for agent in self.get_targets():
            if agent.name not in seen:
                seen.add(agent.name)
                await agent.connections.wait_for_connections(seen)

    def get_targets(
        self, recursive: bool = False, _seen: set[AgentName] | None = None
    ) -> set[AnyAgent[Any, Any]]:
        """Get all currently connected target agents.

        Args:
            recursive: Whether to include targets of targets
        """
        # Get direct targets
        targets = {t for conn in self._connections for t in conn.targets if conn.active}

        if not recursive:
            return targets

        # Track seen agents to prevent cycles
        seen = _seen or {self.owner.name}  # type: ignore
        all_targets = set()

        for target in targets:
            if target.name not in seen:
                _targets = target.connections.get_targets(recursive=True, _seen=seen)
                seen.add(target.name)
                all_targets.add(target)
                all_targets.update(_targets)

        return all_targets

    def has_connection_to(self, target: AnyAgent[Any, Any]) -> bool:
        """Check if target is connected."""
        return any(target in conn.targets for conn in self._connections if conn.active)

    @overload
    def connect_agent_to(
        self,
        other: AnyAgent[Any, Any] | AgentName,
        *,
        connection_type: ConnectionType = "run",
        priority: int = 0,
        delay: timedelta | None = None,
        queued: bool = False,
        queue_strategy: QueueStrategy = "latest",
        transform: AnyTransformFn | None = None,
        filter_condition: AnyFilterFn | None = None,
        stop_condition: AnyFilterFn | None = None,
        exit_condition: AnyFilterFn | None = None,
    ) -> Talk[Any]: ...

    @overload
    def connect_agent_to(
        self,
        other: Team[Any],
        *,
        connection_type: ConnectionType = "run",
        priority: int = 0,
        delay: timedelta | None = None,
        queued: bool = False,
        queue_strategy: QueueStrategy = "latest",
        transform: AnyTransformFn | None = None,
        filter_condition: AnyFilterFn | None = None,
        stop_condition: AnyFilterFn | None = None,
        exit_condition: AnyFilterFn | None = None,
    ) -> TeamTalk: ...

    def connect_agent_to(
        self,
        other: AnyAgent[Any, Any] | Team[Any] | AgentName,
        *,
        connection_type: ConnectionType = "run",
        priority: int = 0,
        delay: timedelta | None = None,
        queued: bool = False,
        queue_strategy: QueueStrategy = "latest",
        transform: AnyTransformFn | None = None,
        filter_condition: AnyFilterFn | None = None,
        stop_condition: AnyFilterFn | None = None,
        exit_condition: AnyFilterFn | None = None,
    ) -> Talk[Any] | TeamTalk:
        """Handle single agent connections."""
        from llmling_agent.agent import Agent, StructuredAgent
        from llmling_agent.delegation.team import Team

        if not isinstance(self.owner, Agent | StructuredAgent):
            msg = "connect_agent_to can only be used with single agents"
            raise TypeError(msg)

        targets = self._resolve_targets(other)
        for target in targets:
            self.agent_connected.emit(target)

        if isinstance(other, Team):
            conns = [
                self.add_connection(
                    self.owner,
                    [target],
                    connection_type=connection_type,
                    priority=priority,
                    delay=delay,
                    queued=queued,
                    queue_strategy=queue_strategy,
                    transform=transform,
                    filter_condition=filter_condition,
                    stop_condition=stop_condition,
                    exit_condition=exit_condition,
                )
                for target in targets
            ]
            return TeamTalk(conns)

        return self.add_connection(
            self.owner,
            targets,
            connection_type=connection_type,
            priority=priority,
            delay=delay,
            queued=queued,
            queue_strategy=queue_strategy,
            transform=transform,
            filter_condition=filter_condition,
            stop_condition=stop_condition,
            exit_condition=exit_condition,
        )

    def add_connection(
        self,
        source: AnyAgent[Any, Any],
        targets: list[AnyAgent[Any, Any]],
        *,
        connection_type: ConnectionType = "run",
        priority: int = 0,
        delay: timedelta | None = None,
        queued: bool = False,
        queue_strategy: QueueStrategy = "latest",
        transform: AnyTransformFn | None = None,
        filter_condition: AnyFilterFn | None = None,
        stop_condition: AnyFilterFn | None = None,
        exit_condition: AnyFilterFn | None = None,
    ) -> Talk[Any]:
        """Add a connection to the manager."""
        connection = Talk(
            source,
            targets,
            connection_type=connection_type,
            priority=priority,
            delay=delay,
            queued=queued,
            queue_strategy=queue_strategy,
            transform=transform,
            filter_condition=filter_condition,
            stop_condition=stop_condition,
            exit_condition=exit_condition,
        )
        self.connection_added.emit(connection)
        self._connections.append(connection)
        return connection

    def connect_group_to(
        self,
        other: AnyAgent[Any, Any] | Team[Any] | AgentName,
        *,
        connection_type: ConnectionType = "run",
        priority: int = 0,
        delay: timedelta | None = None,
        queued: bool = False,
        queue_strategy: QueueStrategy = "latest",
        transform: AnyTransformFn | None = None,
        filter_condition: AnyFilterFn | None = None,
        stop_condition: AnyFilterFn | None = None,
        exit_condition: AnyFilterFn | None = None,
        **kwargs: Any,
    ) -> TeamTalk:
        """Handle group connections."""
        from llmling_agent.delegation.team import Team

        if not isinstance(self.owner, Team):
            msg = "connect_group_to can only be used with agent groups"
            raise TypeError(msg)

        targets = self._resolve_targets(other)
        for target in targets:
            self.agent_connected.emit(target)

        conns = [
            self.add_connection(
                src,
                [t],
                connection_type=connection_type,
                priority=priority,
                delay=delay,
                queued=queued,
                queue_strategy=queue_strategy,
                transform=transform,
                filter_condition=filter_condition,
                stop_condition=stop_condition,
                exit_condition=exit_condition,
            )
            for src in self.owner.agents
            for t in targets
        ]
        return TeamTalk(conns)

    def _resolve_targets(
        self, other: AnyAgent[Any, Any] | Team[Any] | AgentName
    ) -> list[AnyAgent[Any, Any]]:
        """Resolve target(s) to connect to."""
        from llmling_agent.agent import Agent, StructuredAgent
        from llmling_agent.delegation.team import Team

        match other:
            case str():
                if (
                    not isinstance(self.owner, Agent | StructuredAgent)
                    or not self.owner.context.pool
                ):
                    msg = "Pool required for forwarding to agent by name"
                    raise ValueError(msg)
                return [self.owner.context.pool.get_agent(other)]
            case Team():
                return list(other.agents)
        return [other]

    async def trigger_all(self) -> dict[AgentName, list[ChatMessage[Any]]]:
        """Trigger all queued connections."""
        results = {}
        for talk in self._connections:
            if isinstance(talk, Talk) and talk.queued:
                results[talk.source.name] = await talk.trigger()
        return results

    async def trigger_for(
        self, target: AgentName | AnyAgent[Any, Any]
    ) -> list[ChatMessage[Any]]:
        """Trigger queued connections to specific target."""
        target_name = target if isinstance(target, str) else target.name
        results = []
        for talk in self._connections:
            if isinstance(talk, Talk) and talk.queued:  # noqa: SIM102
                if any(t.name == target_name for t in talk.targets):
                    results.extend(await talk.trigger())
        return results

    def disconnect_all(self):
        """Disconnect all managed connections."""
        for conn in self._connections:
            conn.disconnect()
        self._connections.clear()

    def disconnect(self, agent: AnyAgent[Any, Any]):
        """Disconnect a specific agent."""
        to_disconnect: list[Talk | TeamTalk] = []
        for talk in self._connections:
            match talk:
                case Talk():
                    if agent in talk.targets or agent == talk.source:
                        to_disconnect.append(talk)
                case TeamTalk():
                    if agent in talk.targets:
                        to_disconnect.append(talk)

        for talk in to_disconnect:
            talk.active = False
            self._connections.remove(talk)

    async def route_message(self, message: ChatMessage[Any], wait: bool | None = None):
        """Route message to all connections."""
        if wait is not None:
            should_wait = wait
        else:
            should_wait = any(
                self._wait_states.get(t.name, False) for t in self.get_targets()
            )
        msg = "ConnectionManager routing message from %s to %d connections"
        logger.debug(msg, message.name, len(self._connections))

        forwarded_from = [*message.forwarded_from, self.owner.name]  # type: ignore
        message_copy = replace(message, forwarded_from=forwarded_from)

        for talk in self._connections:
            await talk._handle_message(message_copy, None)

        if should_wait:
            await self.wait_for_connections()

    @asynccontextmanager
    async def paused_routing(self):
        """Temporarily pause message routing to connections."""
        active_talks = [talk for talk in self._connections if talk.active]
        for talk in active_talks:
            talk.active = False

        try:
            yield self
        finally:
            for talk in active_talks:
                talk.active = True

    @property
    def stats(self) -> TeamTalkStats:
        """Get aggregated statistics for all connections."""
        return TeamTalkStats(stats=[conn.stats for conn in self._connections])

    def get_connections(self, recursive: bool = False) -> list[Talk[Any]]:
        """Get all Talk connections, flattening TeamTalks."""
        result = []
        seen = set()

        # Get our direct connections
        for conn in self._connections:
            if isinstance(conn, Talk):
                result.append(conn)
            else:  # TeamTalk
                result.extend(t for t in conn if isinstance(t, Talk))

        # Get target connections if recursive
        if recursive:
            for conn in result:
                for target in conn.targets:
                    if target.name not in seen:
                        seen.add(target.name)
                        result.extend(target.connections.get_connections(True))

        return result

    def get_mermaid_diagram(
        self,
        include_details: bool = True,
        recursive: bool = True,
    ) -> str:
        """Generate mermaid flowchart of all connections."""
        lines = ["flowchart LR"]
        connections = self.get_connections(recursive=recursive)

        for talk in connections:
            source = talk.source.name
            for target in talk.targets:
                if include_details:
                    details: list[str] = []
                    details.append(talk.connection_type)
                    if talk.queued:
                        details.append(f"queued({talk.queue_strategy})")
                    if talk._filter_condition:
                        details.append(f"filter:{talk._filter_condition.__name__}")
                    if talk._stop_condition:
                        details.append(f"stop:{talk._stop_condition.__name__}")
                    if talk._exit_condition:
                        details.append(f"exit:{talk._exit_condition.__name__}")
                    elif any([
                        talk._filter_condition,
                        talk._stop_condition,
                        talk._exit_condition,
                    ]):
                        details.append("conditions")

                    label = f"|{' '.join(details)}|" if details else ""
                    lines.append(f"    {source}--{label}-->{target.name}")
                else:
                    lines.append(f"    {source}-->{target.name}")

        return "\n".join(lines)


if __name__ == "__main__":
    from llmling_agent.agent import Agent

    agent = Agent[None]("test_agent")
    agent_2 = Agent[None]("test_agent_2")
    agent_3 = Agent[None]("test_agent_3")
    agent_4 = Agent[None]("test_agent_3")
    _conn_1 = agent >> agent_2
    _conn_2 = agent >> agent_3
    _conn_3 = agent_2 >> agent_4
    print(agent.connections.get_connections(recursive=True))
    print(agent.connections.get_mermaid_diagram())



from __future__ import annotations

import asyncio
import contextlib
from datetime import datetime, timedelta
import io
from typing import TYPE_CHECKING, Any, Literal
from uuid import uuid4

from llmling import ToolError
from pydantic_ai import RunContext  # noqa: TC002

from llmling_agent.models.context import AgentContext  # noqa: TC001


if TYPE_CHECKING:
    from llmling_agent.agent import AnyAgent


async def delegate_to(  # noqa: D417
    ctx: RunContext[AgentContext],
    agent_name: str,
    prompt: str,
) -> str:
    """Delegate a task. Allows to assign a task to task to a specific agent.

    If an action requires you to delegate a task, this tool can be used to assign and
    execute a task. Instructions can be passed via the prompt parameter.

    Args:
        agent_name: The agent to delegate the task to
        prompt: Instructions for the agent to delegate.

    Returns:
        The result of the task you delegated.
    """
    if not ctx.deps.pool:
        msg = "Agent needs to be in a pool to delegate tasks"
        raise ToolError(msg)
    specialist = ctx.deps.pool.get_agent(agent_name)
    result = await specialist.run(prompt)
    return str(result.data)


async def list_available_agents(  # noqa: D417
    ctx: RunContext[AgentContext],
    only_idle: bool = False,
) -> list[str]:
    """List all agents available in the current pool.

    Args:
        only_idle: If True, only returns agents that aren't currently busy.
                    Use this to find agents ready for immediate tasks.

    Returns:
        List of agent names that you can use with delegate_to
    """
    if not ctx.deps.pool:
        msg = "Agent needs to be in a pool to list agents"
        raise ToolError(msg)

    agents = list(ctx.deps.pool.list_agents())
    if only_idle:
        return [n for n in agents if not ctx.deps.pool.get_agent(n).is_busy()]
    return agents


async def create_worker_agent[TDeps](
    ctx: RunContext[AgentContext[TDeps]],
    name: str,
    system_prompt: str,
    model: str | None = None,
) -> str:
    """Create a new agent and register it as a tool.

    The new agent will be available as a tool for delegating specific tasks.
    It inherits the current model unless overridden.
    """
    from llmling_agent import Agent

    if not ctx.deps.pool:
        msg = "Agent needs to be in a pool to list agents"
        raise ToolError(msg)

    model = model or ctx.model.name()
    worker = Agent[TDeps](
        name=name,
        model=model,
        system_prompt=system_prompt,
        context=ctx.deps,
    )
    assert ctx.deps.agent
    tool_info = ctx.deps.agent.register_worker(worker)
    return f"Created worker agent and registered as tool: {tool_info.name}"


async def spawn_delegate[TDeps](
    ctx: RunContext[AgentContext[TDeps]],
    task: str,
    system_prompt: str,
    model: str | None = None,
    capabilities: dict[str, bool] | None = None,
    connect_back: bool = False,
) -> str:
    """Spawn a temporary agent for a specific task.

    Creates an ephemeral agent that will execute the task and clean up automatically
    Optionally connects back to receive results.
    """
    from llmling_agent import Agent

    if not ctx.deps.pool:
        msg = "No agent pool available"
        raise ToolError(msg)

    name = f"delegate_{uuid4().hex[:8]}"
    model = model or ctx.model.name()
    agent = Agent[TDeps](
        name=name,
        model=model,
        system_prompt=system_prompt,
        context=ctx.deps,
    )

    if connect_back:
        assert ctx.deps.agent
        ctx.deps.agent.pass_results_to(agent)

    await agent.run(task)
    return f"Spawned delegate {name} for task"


async def search_history(
    ctx: RunContext[AgentContext],
    query: str | None = None,
    hours: int = 24,
    limit: int = 5,
) -> str:
    """Search conversation history."""
    from llmling_agent_storage.formatters import format_output

    if ctx.deps.capabilities.history_access == "none":
        msg = "No permission to access history"
        raise ToolError(msg)

    provider = ctx.deps.storage.get_history_provider()
    results = await provider.get_filtered_conversations(
        query=query,
        period=f"{hours}h",
        limit=limit,
    )
    return format_output(results)


async def show_statistics(
    ctx: RunContext[AgentContext],
    group_by: Literal["agent", "model", "hour", "day"] = "model",
    hours: int = 24,
) -> str:
    """Show usage statistics for conversations."""
    from llmling_agent_storage.formatters import format_output
    from llmling_agent_storage.models import StatsFilters

    if ctx.deps.capabilities.stats_access == "none":
        msg = "No permission to view statistics"
        raise ToolError(msg)

    cutoff = datetime.now() - timedelta(hours=hours)
    filters = StatsFilters(cutoff=cutoff, group_by=group_by)

    provider = ctx.deps.storage.get_history_provider()
    stats = await provider.get_conversation_stats(filters)

    return format_output(
        {
            "period": f"{hours}h",
            "group_by": group_by,
            "entries": [
                {
                    "name": key,
                    "messages": data["messages"],
                    "total_tokens": data["total_tokens"],
                    "models": sorted(data["models"]),
                }
                for key, data in stats.items()
            ],
        },
        output_format="text",
    )


async def execute_python(ctx: RunContext[AgentContext], code: str) -> str:
    """Execute Python code directly."""
    if not ctx.deps.capabilities.can_execute_code:
        msg = "No permission to execute code"
        raise ToolError(msg)

    try:
        # Capture output
        with io.StringIO() as buf, contextlib.redirect_stdout(buf):
            exec(code, {"__builtins__": __builtins__})
            return buf.getvalue() or "Code executed successfully"
    except Exception as e:  # noqa: BLE001
        return f"Error executing code: {e}"


async def execute_command(ctx: RunContext[AgentContext], command: str) -> str:
    """Execute a shell command."""
    if not ctx.deps.capabilities.can_execute_commands:
        msg = "No permission to execute commands"
        raise ToolError(msg)

    try:
        pipe = asyncio.subprocess.PIPE
        proc = await asyncio.create_subprocess_shell(command, stdout=pipe, stderr=pipe)
        stdout, stderr = await proc.communicate()
        return stdout.decode() or stderr.decode() or "Command completed"
    except Exception as e:  # noqa: BLE001
        return f"Error executing command: {e}"


async def add_agent(  # noqa: D417
    ctx: RunContext[AgentContext],
    name: str,
    system_prompt: str,
    model: str | None = None,
    tools: list[str] | None = None,
    session: str | None = None,
    result_type: str | None = None,
) -> str:
    """Add a new agent to the pool.

    Args:
        name: Name for the new agent
        system_prompt: System prompt defining agent's role/behavior
        model: Optional model override (uses default if not specified)
        tools: Names of tools to enable for this agent
        session: Session ID to recover conversation state from
        result_type: Name of response type from manifest (for structured output)

    Returns:
        Confirmation message about the created agent
    """
    assert ctx.deps.pool, "No agent pool available"
    agent: AnyAgent[Any, Any] = await ctx.deps.pool.add_agent(
        name=name,
        system_prompt=system_prompt,
        model=model,
        tools=tools,
        result_type=result_type,
        session=session,
    )
    return f"Created agent {agent.name} using model {agent.model_name}"


async def ask_agent(  # noqa: D417
    ctx: RunContext[AgentContext],
    agent_name: str,
    message: str,
    *,
    model: str | None = None,
    store_history: bool = True,
) -> str:
    """Send a message to a specific agent and get their response.

    Args:
        agent_name: Name of the agent to interact with
        message: Message to send to the agent
        model: Optional temporary model override
        store_history: Whether to store this exchange in history

    Returns:
        The agent's response
    """
    assert ctx.deps.pool, "No agent pool available"
    agent = ctx.deps.pool.get_agent(agent_name)
    result = await agent.run(
        message,
        model=model,
        store_history=store_history,
    )
    return str(result.content)



"""Function discovery and metadata handling."""

from __future__ import annotations

from dataclasses import dataclass, field
import inspect
from typing import TYPE_CHECKING, Any

from llmling_agent.log import get_logger


if TYPE_CHECKING:
    from collections.abc import Callable, Sequence

    from llmling_agent.common_types import AnyCallable

logger = get_logger(__name__)


@dataclass
class AgentFunction:
    """Metadata for a function that uses agents."""

    func: AnyCallable
    """The actual function to execute."""

    depends_on: list[str] = field(default_factory=list)
    """Names of functions this one depends on."""

    default_inputs: dict[str, Any] = field(default_factory=dict)
    """Default parameter values."""

    name: str = field(init=False)
    """Function name (from function.__name__)."""

    def __post_init__(self):
        """Set name and validate dependencies."""
        self.name = self.func.__name__
        # Extract default inputs from function signature

        sig = inspect.signature(self.func)
        self.default_inputs = {
            name: param.default
            for name, param in sig.parameters.items()
            if param.default is not param.empty
        }
        msg = "Registered agent function %s (deps=%s)"
        logger.debug(msg, self.name, self.depends_on)


def agent_function(
    func: Callable | None = None,
    *,
    depends_on: str | Sequence[str] | None = None,
) -> Callable:
    """Mark a function for automatic agent execution.

    Can be used as simple decorator or with arguments:

    @agent_function
    async def func(): ...

    @agent_function(order=1, depends_on="other_func")
    async def func(): ...

    Args:
        func: Function to mark
        depends_on: Names of functions this one depends on

    Returns:
        Decorated function
    """

    def decorator(func: Callable) -> Callable:
        deps = [depends_on] if isinstance(depends_on, str) else list(depends_on or [])
        metadata = AgentFunction(func=func, depends_on=deps)
        func._agent_function = metadata  # type: ignore
        return func

    return decorator(func) if func is not None else decorator



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, overload

from upath import UPath

from llmling_agent.log import get_logger


if TYPE_CHECKING:
    from llmling_agent.common_types import StrPath


logger = get_logger(__name__)


@overload
async def read_path(path: StrPath, mode: Literal["rt"] = "rt") -> str: ...


@overload
async def read_path(path: StrPath, mode: Literal["rb"]) -> bytes: ...


async def read_path(
    path: StrPath,
    mode: Literal["rt", "rb"] = "rt",
) -> str | bytes:
    """Read file content asynchronously when possible.

    Args:
        path: Path to read
        mode: Read mode ("rt" for text, "rb" for binary)

    Returns:
        File content as string or bytes depending on mode
    """
    import fsspec
    from fsspec.asyn import AsyncFileSystem
    from fsspec.implementations.asyn_wrapper import AsyncFileSystemWrapper
    from morefs.asyn_local import AsyncLocalFileSystem

    path_obj = UPath(path)

    # Try to get native async filesystem first
    if path_obj.protocol in ("", "file"):
        fs = AsyncLocalFileSystem(path_obj.fs)
    else:
        fs = fsspec.filesystem(path_obj.protocol, asynchronous=True)

        if not isinstance(fs, AsyncFileSystem):
            fs = AsyncFileSystemWrapper(path_obj.fs)
    f = await fs.open_async(path_obj.path, mode=mode)
    async with f:
        return await f.read()


if __name__ == "__main__":
    import asyncio

    async def main():
        content = await read_path("README.md")
        print(content)

    asyncio.run(main())



"""Automatic agent function execution."""

from __future__ import annotations

import asyncio
from typing import Any, TYPE_CHECKING

from llmling_agent.delegation import AgentPool
from llmling_agent.running.discovery import agent_function
from llmling_agent.running.executor import discover_functions, execute_functions

if TYPE_CHECKING:
    from llmling_agent.models.agents import AgentsManifest
    from llmling_agent.common_types import StrPath


async def run_agents_async(
    config: StrPath | AgentsManifest,
    *,
    module: str | None = None,
    functions: list[str] | None = None,
    inputs: dict[str, Any] | None = None,
    parallel: bool = False,
) -> dict[str, Any]:
    """Run agent functions asynchronously."""
    # Find functions to run
    if module:
        discovered = discover_functions(module)
    else:
        # Use calling module
        import inspect

        frame = inspect.currentframe()
        while frame:
            if frame.f_globals.get("__name__") != __name__:
                break
            frame = frame.f_back
        if not frame:
            msg = "Could not determine calling module"
            raise RuntimeError(msg)
        discovered = discover_functions(frame.f_globals["__file__"])

    if functions:
        discovered = [f for f in discovered if f.name in functions]

    # Run with pool
    async with AgentPool[None](config) as pool:
        return await execute_functions(
            discovered,
            pool,
            inputs=inputs,
            parallel=parallel,
        )


def run_agents(
    config: str,
    **kwargs: Any,
) -> dict[str, Any]:
    """Run agent functions (sync version)."""
    return asyncio.run(run_agents_async(config, **kwargs))


__all__ = ["agent_function", "run_agents", "run_agents_async"]



"""Message and token usage models."""

from __future__ import annotations

import dataclasses
from dataclasses import asdict, dataclass, field
from datetime import datetime
from typing import Any, Literal, TypedDict
from uuid import uuid4

from pydantic import BaseModel
import tokonomics
from typing_extensions import TypeVar
import yamling

from llmling_agent.common_types import JsonObject, MessageRole  # noqa: TC001
from llmling_agent.log import get_logger
from llmling_agent.models.agents import ToolCallInfo  # noqa: TC001


TContent = TypeVar("TContent", str, BaseModel, default=str)
FormatStyle = Literal["simple", "detailed", "markdown", "custom"]
logger = get_logger(__name__)

SIMPLE_TEMPLATE = """{{ name or role.title() }}: {{ content }}"""

DETAILED_TEMPLATE = """From: {{ name or role.title() }}
Time: {{ timestamp.strftime('%Y-%m-%d %H:%M:%S') }}
----------------------------------------
{{ content }}
----------------------------------------
{%- if show_costs and cost_info %}
Tokens: {{ "{:,}".format(cost_info.token_usage['total']) }}
Cost: ${{ "%.5f"|format(cost_info.total_cost) }}
{%- if response_time %}
Response time: {{ "%.2f"|format(response_time) }}s
{%- endif %}
{%- endif %}
{%- if show_metadata and metadata %}
Metadata:
{%- for key, value in metadata.items() %}
  {{ key }}: {{ value }}
{%- endfor %}
{%- endif %}
{%- if forwarded_from %}
Forwarded via: {{ forwarded_from|join(' -> ') }}
{%- endif %}"""

MARKDOWN_TEMPLATE = """## {{ name or role.title() }}
*{{ timestamp.strftime('%Y-%m-%d %H:%M:%S') }}*

{{ content }}

{%- if show_costs and cost_info %}
---
**Stats:**
- Tokens: {{ "{:,}".format(cost_info.token_usage['total']) }}
- Cost: ${{ "%.4f"|format(cost_info.total_cost) }}
{%- if response_time %}
- Response time: {{ "%.2f"|format(response_time) }}s
{%- endif %}
{%- endif %}

{%- if show_metadata and metadata %}
**Metadata:**
```
{{ metadata|to_yaml }}
```
{%- endif %}

{%- if forwarded_from %}
*Forwarded via: {{ forwarded_from|join(' → ') }}*
{%- endif %}"""

MESSAGE_TEMPLATES = {
    "simple": SIMPLE_TEMPLATE,
    "detailed": DETAILED_TEMPLATE,
    "markdown": MARKDOWN_TEMPLATE,
}


class TokenUsage(TypedDict):
    """Token usage statistics from model responses."""

    total: int
    """Total tokens used"""
    prompt: int
    """Tokens used in the prompt"""
    completion: int
    """Tokens used in the completion"""


@dataclass(frozen=True)
class TokenCost:
    """Combined token and cost tracking."""

    token_usage: TokenUsage
    """Token counts for prompt and completion"""
    total_cost: float
    """Total cost in USD"""

    @classmethod
    async def from_usage(
        cls,
        usage: tokonomics.Usage | None,
        model: str,
        prompt: str,
        completion: str,
    ) -> TokenCost | None:
        """Create result from usage data.

        Args:
            usage: Token counts from model response
            model: Name of the model used
            prompt: The prompt text sent to model
            completion: The completion text received

        Returns:
            TokenCost if usage data available, None otherwise
        """
        if not (
            usage
            and usage.total_tokens is not None
            and usage.request_tokens is not None
            and usage.response_tokens is not None
        ):
            logger.debug("Missing token counts in Usage object")
            return None

        token_usage = TokenUsage(
            total=usage.total_tokens,
            prompt=usage.request_tokens,
            completion=usage.response_tokens,
        )
        logger.debug("Token usage: %s", token_usage)

        cost = await tokonomics.calculate_token_cost(
            model,
            usage.request_tokens,
            usage.response_tokens,
        )
        total_cost = cost.total_cost if cost else 0.0

        return cls(token_usage=token_usage, total_cost=total_cost)


@dataclass
class ChatMessage[TContent]:
    """Common message format for all UI types.

    Generically typed with: ChatMessage[Type of Content]
    The type can either be str or a BaseModel subclass.
    """

    content: TContent
    """Message content, typed as TContent (either str or BaseModel)."""

    role: MessageRole
    """Role of the message sender (user/assistant/system)."""

    model: str | None = None
    """Name of the model that generated this message."""

    metadata: JsonObject = field(default_factory=dict)
    """Additional metadata about the message."""

    timestamp: datetime = field(default_factory=datetime.now)
    """When this message was created."""

    cost_info: TokenCost | None = None
    """Token usage and costs for this specific message if available."""

    message_id: str = field(default_factory=lambda: str(uuid4()))
    """Unique identifier for this message."""

    response_time: float | None = None
    """Time it took the LLM to respond."""

    tool_calls: list[ToolCallInfo] = field(default_factory=list)
    """List of tool calls made during message generation."""

    name: str | None = None
    """Display name for the message sender in UI."""

    forwarded_from: list[str] = field(default_factory=list)
    """List of agent names (the chain) that forwarded this message to the sender."""

    def to_text_message(self) -> ChatMessage[str]:
        """Convert this message to a text-only version."""
        return dataclasses.replace(self, content=str(self.content))  # type: ignore

    def _get_content_str(self) -> str:
        """Get string representation of content."""
        match self.content:
            case str():
                return self.content
            case BaseModel():
                return self.content.model_dump_json(indent=2)
            case _:
                msg = f"Unexpected content type: {type(self.content)}"
                raise ValueError(msg)

    def to_gradio_format(self) -> tuple[str | None, str | None]:
        """Convert to Gradio chatbot format."""
        content_str = self._get_content_str()
        match self.role:
            case "user":
                return (content_str, None)
            case "assistant":
                return (None, content_str)
            case "system":
                return (None, f"System: {content_str}")

    @property
    def data(self) -> TContent:
        """Get content as typed data. Provides compat to RunResult."""
        return self.content

    def format(
        self,
        style: FormatStyle = "simple",
        *,
        template: str | None = None,
        variables: dict[str, Any] | None = None,
        show_metadata: bool = False,
        show_costs: bool = False,
    ) -> str:
        """Format message with configurable style.

        Args:
            style: Predefined style or "custom" for custom template
            template: Custom Jinja template (required if style="custom")
            variables: Additional variables for template rendering
            show_metadata: Whether to include metadata
            show_costs: Whether to include cost information

        Raises:
            ValueError: If style is "custom" but no template provided
                    or if style is invalid
        """
        from jinja2 import Environment

        env = Environment(trim_blocks=True, lstrip_blocks=True)
        env.filters["to_yaml"] = yamling.dump_yaml

        match style:
            case "custom":
                if not template:
                    msg = "Custom style requires a template"
                    raise ValueError(msg)
                template_str = template
            case _ if style in MESSAGE_TEMPLATES:
                template_str = MESSAGE_TEMPLATES[style]
            case _:
                msg = f"Invalid style: {style}"
                raise ValueError(msg)

        template_obj = env.from_string(template_str)
        render_vars = {
            **asdict(self),
            "show_metadata": show_metadata,
            "show_costs": show_costs,
        }
        if variables:
            render_vars.update(variables)

        return template_obj.render(**render_vars)


@dataclass
class Response[TContent]:
    """Response from any source in the agent system."""

    content: ChatMessage[TContent] | str
    """The actual response content (either a ChatMessage or raw text)."""

    source: str
    """Identifies where this response came from (agent/command/tool/stream)."""

    agent_name: str
    """Name of the agent that generated or handled this response."""

    timing: float | None = None
    """Time taken to generate this response in seconds."""

    error: str | None = None
    """Error message if the response generation failed."""

    @property
    def success(self) -> bool:
        """Whether the response was generated successfully."""
        return self.error is None

    @property
    def data(self) -> TContent | str:
        """Direct access to the response content data."""
        return (
            self.content.content
            if isinstance(self.content, ChatMessage)
            else self.content
        )

    def format(
        self,
        style: Literal["simple", "detailed", "markdown"] = "simple",
        *,
        include_context: bool = False,
        **kwargs: Any,
    ) -> str:
        """Format response as string with optional context info."""
        # Get base message formatting
        msg = (
            self.content.format(style, **kwargs)
            if isinstance(self.content, ChatMessage)
            else str(self.content)
        )

        if not include_context:
            return msg

        context_parts = []
        if self.error:
            context_parts.append(f"Error: {self.error}")
        else:
            context_parts.append(f"Source: {self.source}")
            if self.timing:
                context_parts.append(f"Duration: {self.timing:.2f}s")

        return f"{' | '.join(context_parts)}\n{msg}"


@dataclass
class AgentResponse[TResult]:
    """Result from an agent's execution."""

    # TODO: replace with Response

    agent_name: str
    """Name of the agent that produced this result"""

    message: ChatMessage[TResult] | None
    """The actual message with content and metadata"""

    timing: float | None = None
    """Time taken by this agent in seconds"""

    error: str | None = None
    """Error message if agent failed"""

    @property
    def success(self) -> bool:
        """Whether the agent completed successfully."""
        return self.error is None

    @property
    def response(self) -> TResult | None:
        """Convenient access to message content."""
        return self.message.content if self.message else None


class TeamResponse(list[AgentResponse[Any]]):
    """Results from a team execution."""

    def __init__(
        self, responses: list[AgentResponse[Any]], start_time: datetime | None = None
    ):
        super().__init__(responses)
        self.start_time = start_time or datetime.now()
        self.end_time = datetime.now()

    @property
    def duration(self) -> float:
        """Get execution duration in seconds."""
        return (self.end_time - self.start_time).total_seconds()

    @property
    def successful(self) -> list[AgentResponse[Any]]:
        """Get only successful responses."""
        return [r for r in self if r.success]

    @property
    def failed(self) -> list[AgentResponse[Any]]:
        """Get failed responses."""
        return [r for r in self if not r.success]

    def by_agent(self, name: str) -> AgentResponse[Any] | None:
        """Get response from specific agent."""
        return next((r for r in self if r.agent_name == name), None)

    def format_durations(self) -> str:
        """Format execution times."""
        parts = [f"{r.agent_name}: {r.timing:.2f}s" for r in self if r.timing is not None]
        return f"Individual times: {', '.join(parts)}\nTotal time: {self.duration:.2f}s"

    def to_chat_message(self) -> ChatMessage[str]:
        """Convert team response to a single chat message."""
        # Combine all responses into one structured message
        content = "\n\n".join(
            f"[{response.agent_name}]: {response.message.content}"
            for response in self
            if response.message
        )
        meta = {
            "type": "team_response",
            "agents": [r.agent_name for r in self],
            "duration": self.duration,
            "success_count": len(self.successful),
        }
        return ChatMessage(content=content, role="assistant", metadata=meta)  # type: ignore



from __future__ import annotations

from dataclasses import dataclass
import json
from time import perf_counter
from typing import TYPE_CHECKING, Any

from pydantic import BaseModel
from tokonomics import get_available_models
from tokonomics.toko_types import TokenUsage

from llmling_agent.common_types import ModelProtocol
from llmling_agent.log import get_logger
from llmling_agent.models.agents import ToolCallInfo
from llmling_agent.models.content import BaseContent, Content
from llmling_agent.models.messages import ChatMessage, TokenCost
from llmling_agent_providers.base import AgentProvider, ProviderResponse


if TYPE_CHECKING:
    from litellm import ChatCompletionMessageToolCall

    from llmling_agent.tools.base import ToolInfo


logger = get_logger(__name__)


class LiteLLMModel:
    """Compatible model class for LiteLLM."""

    def __init__(self, model_name: str):
        self._name = model_name

    def name(self) -> str:
        return self._name.replace(":", "/")


@dataclass
class LiteLLMRunContext:
    """Simple run context for LiteLLM provider."""

    message_id: str
    model: LiteLLMModel
    prompt: str
    deps: Any


class LiteLLMProvider(AgentProvider[Any]):
    """Provider using LiteLLM for model-agnostic completions."""

    def __init__(
        self,
        name: str,
        model: str | ModelProtocol | None = None,
        *,
        debug: bool = False,
        retries: int = 1,
    ):
        super().__init__(name=name, debug=debug)
        self._model = model
        self.num_retries = retries

    async def get_model_names(self) -> list[str]:
        """Get list of all known model names."""
        return await get_available_models()

    async def handle_tool_call(
        self,
        tool_call: ChatCompletionMessageToolCall,
        tool: ToolInfo,
        message_id: str,
    ) -> tuple[ToolCallInfo, dict]:
        """Handle a single tool call properly."""
        function_args = json.loads(tool_call.function.arguments)
        start_time = perf_counter()
        result = await tool.execute(**function_args)
        info = ToolCallInfo(
            tool_name=tool.name,
            agent_name=self.name,
            args=function_args,
            result=result,
            tool_call_id=tool_call.id,
            timing=perf_counter() - start_time,
            message_id=message_id,
        )
        self.tool_used.emit(info)
        message = {
            "tool_call_id": tool_call.id,
            "role": "tool",
            "name": tool.name,
            "content": result,
        }

        return info, message

    async def generate_response(
        self,
        *prompts: str | Content,
        message_id: str,
        result_type: type[Any] | None = None,
        model: ModelProtocol | str | None = None,
        store_history: bool = True,
        system_prompt: str | None = None,
        **kwargs: Any,
    ) -> ProviderResponse:
        """Generate response using LiteLLM."""
        from litellm import Choices, acompletion
        from litellm.files.main import ModelResponse

        model_name = self._get_model_name(model)
        complete_history = self.conversation.get_history()
        try:
            # Create messages list from history and new prompt
            messages: list[dict[str, Any]] = []
            if system_prompt:
                messages.append({"role": "system", "content": system_prompt})
            if store_history:
                for msg in complete_history:
                    messages.extend(self._convert_message_to_chat(msg))

            # Convert new prompts to message content
            content_parts: list[dict[str, Any]] = []
            for p in prompts:
                match p:
                    case str():
                        content_parts.append({"type": "text", "text": p})
                    case BaseContent():
                        content_parts.append(p.to_openai_format())
            # Add the multi-modal content as user message
            messages.append({"role": "user", "content": content_parts})

            schemas = [t.get_schema() for t in self.tool_manager.get_tools("enabled")]
            # Get completion
            response = await acompletion(
                stream=False,
                model=model_name,
                messages=messages,
                response_format=result_type
                if result_type and issubclass(result_type, BaseModel)
                else None,
                num_retries=self.num_retries,
                tools=schemas or None,
                tool_choice=self.get_tool_choice() if schemas else None,
                **kwargs,
            )
            assert isinstance(response, ModelResponse)
            assert isinstance(response.choices[0], Choices)
            calls: list[ToolCallInfo] = []
            for tool_call in response.choices[0].message.tool_calls or []:
                function_name = tool_call.function.name
                if not function_name:
                    continue
                tool = self.tool_manager.get(function_name)
                info, message = await self.handle_tool_call(tool_call, tool, message_id)
                calls.append(info)
                messages.append(message)
            # Extract content
            content = response.choices[0].message.content
            if content and result_type and issubclass(result_type, BaseModel):
                # Parse JSON string into the requested model
                content = result_type.model_validate_json(content)
            # Create tokonomics usage
            usage = TokenUsage(
                total_tokens=response.usage.prompt_tokens,  # type: ignore
                request_tokens=response.usage.prompt_tokens,  # type: ignore
                response_tokens=response.usage.completion_tokens,  # type: ignore
            )
            cost_and_usage = TokenCost(
                token_usage=usage,
                total_cost=response.usage.cost,  # type: ignore
            )
            # Store in history if requested
            if store_history:
                request_msgs = [ChatMessage(role="user", content=str(p)) for p in prompts]
                response_msg = ChatMessage(role="assistant", content=content)
                self.conversation.add_chat_messages([*request_msgs, response_msg])

            return ProviderResponse(
                content=content,
                tool_calls=calls,
                model_name=model_name,
                cost_and_usage=cost_and_usage,
            )

        except Exception as e:
            logger.exception("LiteLLM completion failed")
            error_msg = f"LiteLLM completion failed: {e}"
            raise RuntimeError(error_msg) from e

    def _get_model_name(self, override: ModelProtocol | str | None = None) -> str:
        """Get effective model name."""
        if isinstance(override, ModelProtocol):
            return override.name()
        if isinstance(override, str):
            return override.replace(":", "/")
        if isinstance(self._model, ModelProtocol):
            return self._model.name()
        if self._model:
            return self._model.replace(":", "/")
        return "openai/gpt-4o-mini"

    def get_tool_choice(self) -> str:
        match self.tool_manager.tool_choice:
            case True:
                return "auto"
            case False:
                return "none"
            case str():
                return self.tool_manager.tool_choice
            case list():
                return "auto"

    def _convert_message_to_chat(self, message: Any) -> list[dict[str, str]]:
        """Convert message to chat format."""
        # This is a basic implementation - would need to properly handle
        # different message types and parts
        return [{"role": "user", "content": str(message)}]

    # async def stream_response(
    #     self,
    #     *prompts: str | Content,
    #     message_id: str,
    #     *,
    #     result_type: type[Any] | None = None,
    #     model: LiteLLMModel | str | None = None,
    #     store_history: bool = True,
    #     system_prompt: str | None = None,

    #     **kwargs: Any,
    # ) -> AbstractAsyncContextManager[StreamingResponseProtocol]:
    #     """Stream response from LiteLLM.

    #     Not implemented yet - would need to handle streaming responses.
    #     """
    #     msg = "Streaming not yet supported"
    #     raise NotImplementedError(msg)


if __name__ == "__main__":
    import logging

    from llmling_agent import Agent

    logging.basicConfig(level=logging.INFO)

    # Create agent with LiteLLM provider
    agent = Agent[Any](provider="litellm", model="openai/gpt-3.5-turbo", name="test")

    # Use run_sync for simple testing
    response = agent.run_sync("Tell me a short joke about Python programming.")

    print(f"\nResponse from {agent.model_name}:")
    print(f"Content: {response.content}")
    if response.cost_info:
        print(f"Tokens: {response.cost_info.token_usage}")
        print(f"Cost: ${response.cost_info.total_cost:.4f}")



"""Tool management for LLMling agents."""

from __future__ import annotations

import asyncio
from collections.abc import Callable, Sequence
from contextlib import AsyncExitStack, contextmanager
from dataclasses import dataclass, field, fields
from datetime import datetime
import os
from typing import TYPE_CHECKING, Any, Literal, Self

from llmling import BaseRegistry, LLMCallableTool, ToolError
from psygnal import Signal

from llmling_agent.log import get_logger
from llmling_agent.mcp_server.client import MCPClient
from llmling_agent.models.mcp_server import MCPServerConfig, SSEMCPServer, StdioMCPServer
from llmling_agent.tools.base import ToolInfo


if TYPE_CHECKING:
    from collections.abc import Iterator
    from types import TracebackType

    from llmling_agent.agent import AnyAgent
    from llmling_agent.common_types import AnyCallable, ToolSource, ToolType
    from llmling_agent.models.context import AgentContext


logger = get_logger(__name__)

MAX_LEN_DESCRIPTION = 1000


class ToolManager(BaseRegistry[str, ToolInfo]):
    """Manages tool registration, enabling/disabling and access.

    Inherits from BaseRegistry providing:
    - Dict-like access: manager["tool_name"] -> ToolInfo
    - Async startup/shutdown: await manager.startup()
    - Event observation: manager.add_observer(observer)
    - Registration: manager.register("tool_name", tool)
    - Listing: manager.list_items()
    - State check: manager.is_empty, manager.has_item()
    - Async iteration: async for name, tool in manager: ...
    """

    @dataclass(frozen=True)
    class ToolStateReset:
        """Emitted when tool states are reset."""

        previous_tools: dict[str, bool]
        new_tools: dict[str, bool]
        timestamp: datetime = field(default_factory=datetime.now)

    tool_states_reset = Signal(ToolStateReset)

    def __init__(
        self,
        tools: Sequence[ToolInfo | ToolType | dict[str, Any]] | None = None,
        *,
        tool_choice: bool | str | list[str] = True,
        context: AgentContext[Any] | None = None,
    ):
        """Initialize tool manager.

        Args:
            tools: Initial tools to register
            tool_choice: Control tool usage:
                - True: Allow all tools
                - False: No tools
                - str: Use specific tool
                - list[str]: Allow specific tools
            context: Tool context
        """
        super().__init__()
        self.tool_choice = tool_choice
        self.context = context
        self._mcp_clients: dict[str, MCPClient] = {}
        self.exit_stack = AsyncExitStack()

        # Register initial tools
        for tool in tools or []:
            t = self._validate_item(tool)
            self.register(t.name, t)

    def __prompt__(self) -> str:
        enabled_tools = [t.name for t in self.values() if t.enabled]
        if not enabled_tools:
            return "No tools available"
        return f"Available tools: {', '.join(enabled_tools)}"

    async def __aenter__(self) -> Self:
        """Enter async context and set up MCP servers."""
        try:
            # Setup MCP servers if configured
            if self.context and self.context.config and self.context.config.mcp_servers:
                await self.setup_mcp_servers(self.context.config.get_mcp_servers())
        except Exception as e:
            # Clean up on error
            await self.__aexit__(type(e), e, e.__traceback__)
            msg = "Failed to initialize tool manager"
            logger.exception(msg, exc_info=e)
            raise RuntimeError(msg) from e
        else:
            return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ):
        """Exit async context."""
        try:
            # Clean up MCP clients through exit stack
            await self.exit_stack.aclose()
            self._mcp_clients.clear()
        except Exception as e:
            msg = "Error during tool manager cleanup"
            logger.exception(msg, exc_info=e)
            raise RuntimeError(msg) from e

    def reset_states(self):
        """Reset all tools to their default enabled states."""
        for info in self.values():
            info.enabled = True

    @property
    def _error_class(self) -> type[ToolError]:
        """Error class for tool operations."""
        return ToolError

    def _validate_item(self, item: ToolInfo | ToolType | dict[str, Any]) -> ToolInfo:
        """Validate and convert items before registration."""
        match item:
            case ToolInfo():
                return item
            case LLMCallableTool():
                return ToolInfo(callable=item)
            case str():
                item = LLMCallableTool.from_callable(item)
                return ToolInfo(callable=item)
            case _ if callable(item):
                tool = LLMCallableTool.from_callable(item)
                return ToolInfo(callable=tool)
            case {"callable": callable_item, **config} if callable(
                callable_item
            ) or isinstance(callable_item, LLMCallableTool):
                # First convert callable to LLMCallableTool if needed
                tool = (
                    callable_item
                    if isinstance(callable_item, LLMCallableTool)
                    else LLMCallableTool.from_callable(callable_item)
                )

                # Get valid fields from ToolInfo dataclass (excluding 'callable')
                valid_keys = {f.name for f in fields(ToolInfo)} - {"callable"}
                tool_config = {k: v for k, v in config.items() if k in valid_keys}

                return ToolInfo(callable=tool, **tool_config)  # type: ignore

            case _:
                typ = type(item)
                msg = f"Item must be ToolInfo, LLMCallableTool, or callable. Got {typ}"
                raise ToolError(msg)

    def enable_tool(self, tool_name: str):
        """Enable a previously disabled tool."""
        if tool_name not in self:
            msg = f"Tool not found: {tool_name}"
            raise ToolError(msg)
        tool_info = self[tool_name]
        tool_info.enabled = True
        self.events.changed(tool_name, tool_info)
        logger.debug("Enabled tool: %s", tool_name)

    def disable_tool(self, tool_name: str):
        """Disable a tool."""
        if tool_name not in self:
            msg = f"Tool not found: {tool_name}"
            raise ToolError(msg)
        tool_info = self[tool_name]
        tool_info.enabled = False
        self.events.changed(tool_name, tool_info)
        logger.debug("Disabled tool: %s", tool_name)

    def is_tool_enabled(self, tool_name: str) -> bool:
        """Check if a tool is currently enabled."""
        return self[tool_name].enabled if tool_name in self else False

    def list_tools(self) -> dict[str, bool]:
        """Get a mapping of all tools and their enabled status."""
        return {name: info.enabled for name, info in self.items()}

    def get_tools(
        self,
        state: Literal["all", "enabled", "disabled"] = "all",
    ) -> list[LLMCallableTool]:
        """Get tool objects based on filters."""
        tools = list(self.values())
        match self.tool_choice:
            case str():
                tools = [self[self.tool_choice]]
            case list():
                tools = [self[name] for name in self.tool_choice]
            case True | None:
                tools = tools
            case _:
                tools = []
        filtered_tools = [info.callable for info in tools if info.matches_filter(state)]
        # Sort by priority (if any have non-default priority)
        if any(self[t.name].priority != 100 for t in filtered_tools):  # noqa: PLR2004
            filtered_tools.sort(key=lambda t: self[t.name].priority)

        return filtered_tools

    def get_tool_names(
        self, state: Literal["all", "enabled", "disabled"] = "all"
    ) -> set[str]:
        """Get tool names based on state."""
        return {name for name, info in self.items() if info.matches_filter(state)}

    def register_tool(
        self,
        tool: LLMCallableTool | AnyCallable | str,
        *,
        name_override: str | None = None,
        description_override: str | None = None,
        enabled: bool = True,
        source: ToolSource = "runtime",
        priority: int = 100,
        requires_confirmation: bool = False,
        requires_capability: str | None = None,
        metadata: dict[str, str] | None = None,
    ) -> ToolInfo:
        """Register a new tool with custom settings.

        Args:
            tool: Tool to register (callable, LLMCallableTool, or import path)
            enabled: Whether tool is initially enabled
            name_override: Optional name override for the tool
            description_override: Optional description override for the tool
            source: Tool source (runtime/agent/builtin/dynamic)
            priority: Execution priority (lower = higher priority)
            requires_confirmation: Whether tool needs confirmation
            requires_capability: Optional capability needed to use tool
            metadata: Additional tool metadata

        Returns:
            Created ToolInfo instance
        """
        # First convert to basic ToolInfo
        match tool:
            case LLMCallableTool():
                llm_tool = tool
                llm_tool.name = name_override or llm_tool.name
                llm_tool.description = description_override or llm_tool.description
            case _:
                llm_tool = LLMCallableTool.from_callable(
                    tool,
                    name_override=name_override,
                    description_override=description_override,
                )

        if llm_tool.description and len(llm_tool.description) > MAX_LEN_DESCRIPTION:
            msg = f"Too long description for {tool}"
            raise ToolError(msg)
        tool_info = ToolInfo(
            llm_tool,
            enabled=enabled,
            source=source,
            priority=priority,
            requires_confirmation=requires_confirmation,
            requires_capability=requires_capability,
            metadata=metadata or {},
        )
        # Register the tool
        self.register(tool_info.name, tool_info)
        return tool_info

    def register_worker(
        self,
        worker: AnyAgent[Any, Any],
        *,
        name: str | None = None,
        reset_history_on_run: bool = True,
        pass_message_history: bool = False,
        share_context: bool = False,
        parent: AnyAgent[Any, Any] | None = None,
    ) -> ToolInfo:
        """Register an agent as a worker tool.

        Args:
            worker: Agent to register as worker
            name: Optional name override for the worker tool
            reset_history_on_run: Whether to clear history before each run
            pass_message_history: Whether to pass parent's message history
            share_context: Whether to pass parent's context/deps
            parent: Optional parent agent for history/context sharing
        """
        tool = worker.to_agent_tool(
            parent=parent,
            name=name,
            reset_history_on_run=reset_history_on_run,
            pass_message_history=pass_message_history,
            share_context=share_context,
        )
        msg = "Registering worker %s as tool %s"
        logger.debug(msg, worker.name, tool.name)
        return self.register_tool(tool, source="agent", metadata={"agent": worker.name})

    def reset(self):
        """Reset tool states."""
        old_tools = self.list_tools()
        self.reset_states()
        new_tools = self.list_tools()

        event = self.ToolStateReset(old_tools, new_tools)
        self.tool_states_reset.emit(event)

    async def setup_mcp_servers(self, servers: list[MCPServerConfig]):
        """Set up multiple MCP server integrations.

        Args:
            servers: List of MCP server configurations
        """
        try:
            for server in servers:
                if not server.enabled:
                    continue

                match server:
                    case StdioMCPServer():
                        # Create client with stdio mode
                        client = MCPClient(stdio_mode=True)
                        client = await self.exit_stack.enter_async_context(client)
                        env = os.environ.copy()
                        # Update with server-specific environment if provided
                        if server.environment:
                            env.update(server.environment)
                        # Ensure UTF-8 encoding
                        env["PYTHONIOENCODING"] = "utf-8"
                        await client.connect(server.command, args=server.args, env=env)
                        # Store client
                        client_id = f"{server.command}_{' '.join(server.args)}"
                        self._mcp_clients[client_id] = client

                        # Register tools
                        self.register_mcp_tools(client)

                    case SSEMCPServer():
                        # SSE client without stdio mode
                        client = MCPClient(stdio_mode=False)
                        client = await self.exit_stack.enter_async_context(client)

                        await client.connect(
                            command="",  # Not used for SSE
                            args=[],  # Not used for SSE
                            url=server.url,
                            env=server.environment,
                        )

                        # Store client
                        client_id = f"sse_{server.url}"
                        self._mcp_clients[client_id] = client

                        # Register tools
                        self.register_mcp_tools(client)

        except Exception as e:
            msg = "Failed to setup MCP servers"
            logger.exception(msg, exc_info=e)
            raise RuntimeError(msg) from e

    async def cleanup(self):
        """Clean up resources including all MCP clients."""
        try:
            # Store tools to remove
            to_remove = [
                name
                for name, info in self.items()
                if info.metadata.get("mcp_tool") is not None
            ]

            try:
                # Clean up exit stack (which includes MCP clients)
                await self.exit_stack.aclose()
            except RuntimeError as e:
                if "different task" in str(e):
                    # Handle task context mismatch
                    current_task = asyncio.current_task()
                    if current_task:
                        loop = asyncio.get_running_loop()
                        await loop.create_task(self.exit_stack.aclose())
                else:
                    raise

            self._mcp_clients.clear()

            # Remove MCP tools
            for name in to_remove:
                del self._items[name]

        except Exception as e:
            msg = "Error during tool manager cleanup"
            logger.exception(msg, exc_info=e)
            raise RuntimeError(msg) from e

    @contextmanager
    def temporary_tools(
        self,
        tools: ToolType | Sequence[ToolType],
        *,
        exclusive: bool = False,
    ) -> Iterator[list[ToolInfo]]:
        """Temporarily register tools.

        Args:
            tools: Tool(s) to register
            exclusive: Whether to temporarily disable all other tools

        Yields:
            List of registered tool infos

        Example:
            ```python
            with tool_manager.temporary_tools([tool1, tool2], exclusive=True) as tools:
                # Only tool1 and tool2 are available
                await agent.run(prompt)
            # Original tool states are restored
            ```
        """
        # Normalize inputs to lists
        tools_list: list[ToolType] = (
            [tools] if not isinstance(tools, Sequence) else list(tools)
        )

        # Store original tool states if exclusive
        original_states: dict[str, bool] = {}
        if exclusive:
            original_states = {name: tool.enabled for name, tool in self.items()}
            # Disable all existing tools
            for t in self.values():
                t.enabled = False

        # Register all tools
        registered_tools: list[ToolInfo] = []
        try:
            for tool in tools_list:
                tool_info = self.register_tool(tool)
                registered_tools.append(tool_info)
            yield registered_tools

        finally:
            # Remove temporary tools
            for tool_info in registered_tools:
                del self[tool_info.name]

            # Restore original tool states if exclusive
            if exclusive:
                for name_, was_enabled in original_states.items():
                    if t := self.get(name_):
                        t.enabled = was_enabled

    def register_mcp_tools(
        self,
        mcp_client: MCPClient,
    ) -> list[ToolInfo]:
        """Register MCP tools with tool manager."""
        registered = []

        for mcp_tool in mcp_client._available_tools:
            # Create properly typed callable from schema
            tool_callable = mcp_client.create_tool_callable(mcp_tool)

            # The function already has proper typing, so no schema override needed
            llm_tool = LLMCallableTool.from_callable(tool_callable)

            metadata = {"mcp_tool": mcp_tool.name}
            tool_info = self.register_tool(llm_tool, source="mcp", metadata=metadata)
            registered.append(tool_info)

            logger.debug("Registered MCP tool: %s", mcp_tool.name)

        return registered

    def tool(
        self,
        name: str | None = None,
        *,
        description: str | None = None,
        enabled: bool = True,
        source: ToolSource = "runtime",
        priority: int = 100,
        requires_confirmation: bool = False,
        requires_capability: str | None = None,
        metadata: dict[str, str] | None = None,
    ) -> Callable[[AnyCallable], AnyCallable]:
        """Decorator to register a function as a tool.

        Args:
            name: Optional override for tool name (defaults to function name)
            description: Optional description override
            enabled: Whether tool is initially enabled
            source: Tool source type
            priority: Execution priority (lower = higher)
            requires_confirmation: Whether tool needs confirmation
            requires_capability: Optional required capability
            metadata: Additional tool metadata

        Returns:
            Decorator function that registers the tool

        Example:
            @tool_manager.register(
                name="search_docs",
                description="Search documentation",
                requires_confirmation=True
            )
            async def search(query: str) -> str:
                '''Search the docs.'''
                return "Results..."
        """

        def decorator(func: AnyCallable) -> AnyCallable:
            self.register_tool(
                func,
                name_override=name,
                description_override=description,
                enabled=enabled,
                source=source,
                priority=priority,
                requires_confirmation=requires_confirmation,
                requires_capability=requires_capability,
                metadata=metadata,
            )
            return func

        return decorator



"""SQLModel-based storage provider."""

from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any

from sqlalchemy import JSON, Column, Engine, and_, or_
from sqlalchemy.sql import expression
from sqlmodel import Session, SQLModel, desc, select

from llmling_agent.log import get_logger
from llmling_agent.models.messages import ChatMessage, TokenCost
from llmling_agent.utils.parse_time import parse_time_period
from llmling_agent_storage.base import StorageProvider
from llmling_agent_storage.models import (
    ConversationData,
    QueryFilters,
    StatsFilters,
)
from llmling_agent_storage.sql_provider.models import Conversation, Message


if TYPE_CHECKING:
    from collections.abc import Sequence

    from sqlmodel.sql.expression import SelectOfScalar
    from tokonomics.toko_types import TokenUsage

    from llmling_agent.common_types import JsonValue
    from llmling_agent.models.agents import ToolCallInfo
    from llmling_agent.models.session import SessionQuery
    from llmling_agent.models.storage import SQLStorageConfig


logger = get_logger(__name__)


class SQLModelProvider(StorageProvider[Message]):
    """Storage provider using SQLModel.

    Can work with any database supported by SQLAlchemy/SQLModel.
    Provides efficient SQL-based filtering and storage.
    """

    can_load_history = True

    def __init__(
        self,
        config: SQLStorageConfig,
        engine: Engine,
    ):
        """Initialize provider with database engine.

        Args:
            config: Configuration for provider
            engine: SQLModel engine instance
        """
        from llmling_agent_storage.sql_provider.models import SQLModel

        super().__init__(config)
        self.engine = engine
        SQLModel.metadata.create_all(self.engine)
        self._init_database(auto_migrate=config.auto_migration)

    def _init_database(self, auto_migrate: bool = True):
        """Initialize database tables and optionally migrate columns.

        Args:
            auto_migrate: Whether to automatically add missing columns
        """
        from sqlalchemy import inspect
        from sqlalchemy.sql import text

        from llmling_agent_storage.sql_provider.models import SQLModel

        # Create tables if they don't exist
        SQLModel.metadata.create_all(self.engine)

        # Optionally add missing columns
        if auto_migrate:
            with self.engine.connect() as conn:
                inspector = inspect(self.engine)

                # For each table in our models
                for table_name, table in SQLModel.metadata.tables.items():
                    existing = {col["name"] for col in inspector.get_columns(table_name)}

                    # For each column in model that doesn't exist in DB
                    for col in table.columns:
                        if col.name not in existing:
                            # Create ALTER TABLE statement based on column type
                            type_sql = col.type.compile(self.engine.dialect)
                            nullable = "" if col.nullable else " NOT NULL"
                            default = self._get_column_default(col)
                            sql = f"ALTER TABLE {table_name} ADD COLUMN {col.name} {type_sql}{nullable}{default}"  # noqa: E501
                            conn.execute(text(sql))

                conn.commit()

    @staticmethod
    def _get_column_default(column: Any) -> str:
        """Get SQL DEFAULT clause for column."""
        if column.default is None:
            return ""
        if hasattr(column.default, "arg"):
            # Simple default value
            return f" DEFAULT {column.default.arg}"
        if hasattr(column.default, "sqltext"):
            # Computed default
            return f" DEFAULT {column.default.sqltext}"
        return ""

    def cleanup(self):
        """Clean up database resources."""
        self.engine.dispose()

    async def filter_messages(self, query: SessionQuery) -> list[ChatMessage[str]]:
        """Filter messages using SQL queries."""
        with Session(self.engine) as session:
            stmt = self._build_message_query(query)
            messages = session.exec(stmt).all()
            return [self._to_chat_message(msg) for msg in messages]

    async def log_message(
        self,
        *,
        conversation_id: str,
        content: str,
        role: str,
        name: str | None = None,
        cost_info: TokenCost | None = None,
        model: str | None = None,
        response_time: float | None = None,
        forwarded_from: list[str] | None = None,
    ):
        """Log message to database."""
        from llmling_agent_storage.sql_provider.models import Message

        provider, model_name = self._parse_model_info(model)

        with Session(self.engine) as session:
            msg = Message(
                conversation_id=conversation_id,
                content=content,
                role=role,
                name=name,
                model=model,
                model_provider=provider,
                model_name=model_name,
                response_time=response_time,
                total_tokens=cost_info.token_usage["total"] if cost_info else None,
                prompt_tokens=cost_info.token_usage["prompt"] if cost_info else None,
                completion_tokens=cost_info.token_usage["completion"]
                if cost_info
                else None,
                cost=cost_info.total_cost if cost_info else None,
                forwarded_from=forwarded_from,
                timestamp=datetime.now(),
            )
            session.add(msg)
            session.commit()

    async def log_conversation(
        self,
        *,
        conversation_id: str,
        agent_name: str,
        start_time: datetime | None = None,
    ):
        """Log conversation to database."""
        from llmling_agent_storage.sql_provider.models import Conversation

        with Session(self.engine) as session:
            conversation = Conversation(
                id=conversation_id,
                agent_name=agent_name,
                start_time=start_time or datetime.now(),
            )
            session.add(conversation)
            session.commit()

    async def log_tool_call(
        self,
        *,
        conversation_id: str,
        message_id: str,
        tool_call: ToolCallInfo,
    ):
        """Log tool call to database."""
        from llmling_agent_storage.sql_provider.models import ToolCall

        with Session(self.engine) as session:
            call = ToolCall(
                conversation_id=conversation_id,
                message_id=message_id,
                tool_call_id=tool_call.tool_call_id,
                timestamp=tool_call.timestamp,
                tool_name=tool_call.tool_name,
                args=tool_call.args,
                result=str(tool_call.result),
            )
            session.add(call)
            session.commit()

    async def log_command(
        self,
        *,
        agent_name: str,
        session_id: str,
        command: str,
        context_type: type | None = None,
        metadata: dict[str, JsonValue] | None = None,
    ):
        """Log command to database."""
        from llmling_agent_storage.sql_provider.models import CommandHistory

        with Session(self.engine) as session:
            history = CommandHistory(
                session_id=session_id,
                agent_name=agent_name,
                command=command,
                context_type=context_type.__name__ if context_type else None,
                context_metadata=metadata or {},
            )
            session.add(history)
            session.commit()

    async def get_filtered_conversations(
        self,
        agent_name: str | None = None,
        period: str | None = None,
        since: datetime | None = None,
        query: str | None = None,
        model: str | None = None,
        limit: int | None = None,
        *,
        compact: bool = False,
        include_tokens: bool = False,
    ) -> list[ConversationData]:
        """Get filtered conversations with formatted output."""
        # Convert period to since if provided
        if period:
            since = datetime.now() - parse_time_period(period)

        # Create filters
        filters = QueryFilters(
            agent_name=agent_name,
            since=since,
            query=query,
            model=model,
            limit=limit,
        )

        # Use existing get_conversations method
        conversations = await self.get_conversations(filters)
        return [
            self._format_conversation(
                conv, msgs, compact=compact, include_tokens=include_tokens
            )
            for conv, msgs in conversations
        ]

    async def get_commands(
        self,
        agent_name: str,
        session_id: str,
        *,
        limit: int | None = None,
        current_session_only: bool = False,
    ) -> list[str]:
        """Get command history from database."""
        from llmling_agent_storage.sql_provider.models import CommandHistory

        with Session(self.engine) as session:
            query = select(CommandHistory)
            if current_session_only:
                query = query.where(CommandHistory.session_id == str(session_id))
            else:
                query = query.where(CommandHistory.agent_name == agent_name)

            query = query.order_by(desc(CommandHistory.timestamp))  # type: ignore
            if limit:
                query = query.limit(limit)

            return [h.command for h in session.exec(query)]

    def _build_message_query(
        self,
        query: SessionQuery,
    ) -> SelectOfScalar:
        """Build SQLModel query from SessionQuery."""
        from llmling_agent_storage.sql_provider.models import Message

        stmt = select(Message).order_by(Message.timestamp)  # type: ignore

        conditions: list[Any] = []
        if query.name:
            conditions.append(Message.conversation_id == query.name)
        if query.agents:
            agent_conditions = [Column("name").in_(query.agents)]
            if query.include_forwarded:
                agent_conditions.append(
                    and_(
                        Column("forwarded_from").isnot(None),
                        expression.cast(Column("forwarded_from"), JSON).contains(
                            list(query.agents)
                        ),  # type: ignore
                    )
                )
            conditions.append(or_(*agent_conditions))
        if query.since and (cutoff := query.get_time_cutoff()):
            conditions.append(Message.timestamp >= cutoff)
        if query.until:
            conditions.append(Message.timestamp <= datetime.fromisoformat(query.until))
        if query.contains:
            conditions.append(Message.content.contains(query.contains))  # type: ignore
        if query.roles:
            conditions.append(Message.role.in_(query.roles))  # type: ignore

        if conditions:
            stmt = stmt.where(and_(*conditions))
        if query.limit:
            stmt = stmt.limit(query.limit)

        return stmt  # type: ignore

    def _to_chat_message(self, db_message: Message) -> ChatMessage[str]:
        """Convert database message to ChatMessage."""
        cost_info = None
        if db_message.total_tokens is not None:
            cost_info = TokenCost(
                token_usage={
                    "total": db_message.total_tokens or 0,
                    "prompt": db_message.prompt_tokens or 0,
                    "completion": db_message.completion_tokens or 0,
                },
                total_cost=db_message.cost or 0.0,
            )

        return ChatMessage[str](
            content=db_message.content,
            role=db_message.role,  # type: ignore
            name=db_message.name,
            model=db_message.model,
            cost_info=cost_info,
            response_time=db_message.response_time,
            forwarded_from=db_message.forwarded_from or [],
        )

    @staticmethod
    def _parse_model_info(model: str | None) -> tuple[str | None, str | None]:
        """Parse model string into provider and name."""
        if not model:
            return None, None

        # Try splitting by ':' or '/'
        parts = model.split(":") if ":" in model else model.split("/")

        if len(parts) == 2:  # noqa: PLR2004
            provider, name = parts
            return provider.lower(), name

        # No provider specified, try to infer
        name = parts[0]
        if name.startswith(("gpt-", "text-", "dall-e")):
            return "openai", name
        if name.startswith("claude"):
            return "anthropic", name
        if name.startswith(("llama", "mistral")):
            return "meta", name

        return None, name

    async def get_conversations(
        self,
        filters: QueryFilters,
    ) -> list[tuple[ConversationData, Sequence[ChatMessage[str]]]]:
        """Get filtered conversations using SQL queries."""
        with Session(self.engine) as session:
            results: list[tuple[ConversationData, Sequence[ChatMessage[str]]]] = []

            # Base conversation query
            conv_query = select(Conversation)

            if filters.agent_name:
                conv_query = conv_query.where(
                    Conversation.agent_name == filters.agent_name
                )

            if filters.since:
                # Changed: Gets conversations that STARTED after the cutoff time
                conv_query = conv_query.where(Conversation.start_time >= filters.since)

            conv_query = conv_query.order_by(desc(Conversation.start_time))  # type: ignore
            if filters.limit:
                conv_query = conv_query.limit(filters.limit)

            conversations = session.exec(conv_query).all()

            for conv in conversations:
                msg_query = select(Message).where(Message.conversation_id == conv.id)

                if filters.query:
                    msg_query = msg_query.where(Message.content.contains(filters.query))  # type: ignore
                if filters.model:
                    msg_query = msg_query.where(Message.model_name == filters.model)

                msg_query = msg_query.order_by(Message.timestamp.asc())  # type: ignore
                messages = session.exec(msg_query).all()

                if not messages:
                    continue

                chat_messages = [self._to_chat_message(msg) for msg in messages]
                conv_data = self._format_conversation(conv, messages)
                results.append((conv_data, chat_messages))

            return results

    async def get_conversation_stats(
        self,
        filters: StatsFilters,
    ) -> dict[str, dict[str, Any]]:
        """Get statistics using SQL aggregations."""
        from llmling_agent_storage.sql_provider.models import Conversation, Message

        with Session(self.engine) as session:
            # Base query for stats
            query = (
                select(  # type: ignore[call-overload]
                    Message.model,
                    Conversation.agent_name,
                    Message.timestamp,
                    Message.total_tokens,
                    Message.prompt_tokens,
                    Message.completion_tokens,
                )
                .join(Conversation, Message.conversation_id == Conversation.id)
                .where(Message.timestamp > filters.cutoff)
            )

            if filters.agent_name:
                query = query.where(Conversation.agent_name == filters.agent_name)

            # Execute query and get raw data
            rows = [
                (
                    model,
                    agent,
                    timestamp,
                    TokenCost(
                        token_usage={
                            "total": total or 0,
                            "prompt": prompt or 0,
                            "completion": completion or 0,
                        },
                        total_cost=0.0,  # We don't store this in DB
                    )
                    if total or prompt or completion
                    else None,
                )
                for model, agent, timestamp, total, prompt, completion in session.exec(
                    query
                )
            ]

            # Use base class aggregation
            return self.aggregate_stats(rows, filters.group_by)

    @staticmethod
    def _aggregate_token_usage(
        messages: Sequence[Message | ChatMessage[str]],
    ) -> TokenUsage:
        """Sum up tokens from a sequence of messages."""
        total = prompt = completion = 0
        for msg in messages:
            if isinstance(msg, Message):
                total += msg.total_tokens or 0
                prompt += msg.prompt_tokens or 0
                completion += msg.completion_tokens or 0
            elif msg.cost_info:
                total += msg.cost_info.token_usage.get("total", 0)
                prompt += msg.cost_info.token_usage.get("prompt", 0)
                completion += msg.cost_info.token_usage.get("completion", 0)
        return {"total": total, "prompt": prompt, "completion": completion}

    async def reset(
        self,
        *,
        agent_name: str | None = None,
        hard: bool = False,
    ) -> tuple[int, int]:
        """Reset database storage."""
        from sqlalchemy import text

        from llmling_agent_storage.sql_provider.queries import (
            DELETE_AGENT_CONVERSATIONS,
            DELETE_AGENT_MESSAGES,
            DELETE_ALL_CONVERSATIONS,
            DELETE_ALL_MESSAGES,
        )

        with Session(self.engine) as session:
            if hard:
                if agent_name:
                    msg = "Hard reset cannot be used with agent_name"
                    raise ValueError(msg)
                # Drop and recreate all tables
                SQLModel.metadata.drop_all(self.engine)
                session.commit()
                # Recreate schema
                self._init_database()
                return 0, 0

            # Get counts first
            conv_count, msg_count = await self.get_conversation_counts(
                agent_name=agent_name
            )

            # Delete data
            if agent_name:
                session.execute(text(DELETE_AGENT_MESSAGES), {"agent": agent_name})
                session.execute(text(DELETE_AGENT_CONVERSATIONS), {"agent": agent_name})
            else:
                session.execute(text(DELETE_ALL_MESSAGES))
                session.execute(text(DELETE_ALL_CONVERSATIONS))

            session.commit()
            return conv_count, msg_count

    async def get_conversation_counts(
        self,
        *,
        agent_name: str | None = None,
    ) -> tuple[int, int]:
        """Get conversation and message counts."""
        from llmling_agent_storage.sql_provider import Conversation, Message

        with Session(self.engine) as session:
            if agent_name:
                conv_query = select(Conversation).where(
                    Conversation.agent_name == agent_name
                )
                msg_query = (
                    select(Message)
                    .join(Conversation)
                    .where(Conversation.agent_name == agent_name)
                )
            else:
                conv_query = select(Conversation)
                msg_query = select(Message)

            conv_count = len(session.exec(conv_query).all())
            msg_count = len(session.exec(msg_query).all())

            return conv_count, msg_count

    def _format_conversation(
        self,
        conv: Conversation | ConversationData,
        messages: Sequence[Message | ChatMessage[str]],
        *,
        include_tokens: bool = False,
        compact: bool = False,
    ) -> ConversationData:
        """Format SQL conversation model to ConversationData."""
        msgs = list(messages)
        if compact and len(msgs) > 1:
            msgs = [msgs[0], msgs[-1]]

        # Convert both Conversation and ConversationData to dict format
        if isinstance(conv, Conversation):
            conv_dict = {
                "id": conv.id,
                "agent": conv.agent_name,
                "start_time": conv.start_time.isoformat(),
            }
        else:
            conv_dict = {
                "id": conv["id"],
                "agent": conv["agent"],
                "start_time": conv["start_time"],
            }

        # Convert messages to ChatMessage format if needed
        chat_messages = [
            msg if isinstance(msg, ChatMessage) else self._to_chat_message(msg)
            for msg in msgs
        ]

        return ConversationData(
            id=conv_dict["id"],
            agent=conv_dict["agent"],
            start_time=conv_dict["start_time"],
            messages=[
                {
                    "role": msg.role,
                    "content": msg.content,
                    "timestamp": msg.timestamp.isoformat(),
                    "model": msg.model,
                    "name": msg.name,
                    "token_usage": msg.cost_info.token_usage if msg.cost_info else None,
                    "cost": msg.cost_info.total_cost if msg.cost_info else None,
                    "response_time": msg.response_time,
                }
                for msg in chat_messages
            ],
            token_usage=self._aggregate_token_usage(messages) if include_tokens else None,
        )



"""Core data models for LLMling agent."""

from llmling_agent.models.agents import AgentsManifest, AgentConfig
from llmling_agent.models.messages import (
    ChatMessage,
    TokenUsage,
    TokenCost,
    AgentResponse,
    TeamResponse,
)
from llmling_agent.models.resources import ResourceInfo
from llmling_agent.models.context import AgentContext
from llmling_agent.models.forward_targets import ForwardingTarget
from llmling_agent.models.session import SessionQuery
from llmling_agent.models.mcp_server import (
    MCPServerBase,
    StdioMCPServer,
    MCPServerConfig,
    SSEMCPServer,
)

__all__ = [
    "AgentConfig",
    "AgentContext",
    "AgentResponse",
    "AgentsManifest",
    "ChatMessage",
    "ForwardingTarget",
    "MCPServerBase",
    "MCPServerConfig",
    "ResourceInfo",
    "SSEMCPServer",
    "SessionQuery",
    "StdioMCPServer",
    "TeamResponse",
    "TokenCost",
    "TokenUsage",
]
